# Copyright Red Hat
#
# snapm/_snapm.py - Snapshot Manager global definitions
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
"""
Global definitions for the top-level snapm package.
"""
from uuid import UUID, uuid5
from datetime import datetime
from enum import Enum
import string
import json
import math
import logging
import re
import os

_log = logging.getLogger(__name__)

_log_debug = _log.debug
_log_info = _log.info
_log_warn = _log.warning
_log_error = _log.error

# Snapm debugging mask
SNAPM_DEBUG_MANAGER = 1
SNAPM_DEBUG_COMMAND = 2
SNAPM_DEBUG_REPORT = 4
SNAPM_DEBUG_ALL = SNAPM_DEBUG_MANAGER | SNAPM_DEBUG_COMMAND | SNAPM_DEBUG_REPORT

__DEBUG_MASK = 0


NAMESPACE_SNAPSHOT_SET = UUID("{952f0e38-24a1-406d-adf6-0e9fb3c707d8}")
NAMESPACE_SNAPSHOT = UUID("{c17d07c7-1482-43b7-9b3c-12d490622d93}")

_SIZE_RE = re.compile(r"^(?P<size>[0-9]+)(?P<units>([KMGTPEZkmgtpez]i{,1})?[Bb]{,1})$")

#: All suffixes are expressed in powers of two.
_SIZE_SUFFIXES = {
    "B": 1,
    "K": 2**10,
    "M": 2**20,
    "G": 2**30,
    "T": 2**40,
    "P": 2**50,
    "E": 2**60,
    "Z": 2**70,
}

SECTOR_SIZE = 512

ETC_FSTAB = "/etc/fstab"

DEFAULT_PERCENT_USED = 200.0
DEFAULT_PERCENT_SIZE = 25.0

# Constants for SnapshotSet property names
SNAPSET_NAME = "SnapsetName"
SNAPSET_SOURCES = "Sources"
SNAPSET_MOUNT_POINTS = "MountPoints"
SNAPSET_DEVICES = "Devices"
SNAPSET_NR_SNAPSHOTS = "NrSnapshots"
SNAPSET_TIME = "Time"
SNAPSET_TIMESTAMP = "Timestamp"
SNAPSET_UUID = "UUID"
SNAPSET_STATUS = "Status"
SNAPSET_AUTOACTIVATE = "Autoactivate"
SNAPSET_BOOTABLE = "Bootable"
SNAPSET_BOOT_ENTRIES = "BootEntries"
SNAPSET_SNAPSHOT_ENTRY = "SnapshotEntry"
SNAPSET_REVERT_ENTRY = "RevertEntry"
SNAPSET_SNAPSHOTS = "Snapshots"

# Constants for Snapshot property names
SNAPSHOT_NAME = "Name"
SNAPSHOT_ORIGIN = "Origin"
SNAPSHOT_SOURCE = "Source"
SNAPSHOT_MOUNT_POINT = "MountPoint"
SNAPSHOT_PROVIDER = "Provider"
SNAPSHOT_UUID = "UUID"
SNAPSHOT_STATUS = "Status"
SNAPSHOT_SIZE = "Size"
SNAPSHOT_FREE = "Free"
SNAPSHOT_SIZE_BYTES = "SizeBytes"
SNAPSHOT_FREE_BYTES = "FreeBytes"
SNAPSHOT_AUTOACTIVATE = "Autoactivate"
SNAPSHOT_DEV_PATH = "DevicePath"


# Constant for allow-listed name characters
SNAPM_VALID_NAME_CHARS = set(
    string.ascii_lowercase + string.ascii_uppercase + string.digits + "+_.-"
)


class SnapmLogger(logging.Logger):
    """
    Snapm logging wrapper class: wrap the Logger.debug() method
    to allow filtering of submodule debug messages by log mask.
    """

    mask_bits = 0

    def set_debug_mask(self, mask_bits):
        """
        Set the debug mask for this ``SnapmLogger``.

        This should normally be set to the ``SNAPM_DEBUG_*`` value
        corresponding to the ``snapm`` sub-module that this instance
        of ``SnapmLogger`` belongs to.

        :param mask_bits: The bits to set in this logger's mask.
        :rtype: None
        """
        if mask_bits < 0 or mask_bits > SNAPM_DEBUG_ALL:
            raise ValueError(
                f"Invalid SnapmLogger mask bits: 0x{(mask_bits & ~SNAPM_DEBUG_ALL):x}"
            )

        self.mask_bits = mask_bits

    def debug_masked(self, msg, *args, **kwargs):
        """
        Log a debug message if it passes the current debug mask.

        Log the specified message if it passes the current logger
        debug mask.

        :param msg: the message to be logged
        :rtype: None
        """
        if self.mask_bits & get_debug_mask():
            self.debug(msg, *args, **kwargs)


logging.setLoggerClass(SnapmLogger)


def get_debug_mask():
    """
    Return the current debug mask for the ``snapm`` package.

    :returns: The current debug mask value
    :rtype: int
    """
    return __DEBUG_MASK


def set_debug_mask(mask):
    """
    Set the debug mask for the ``snapm`` package.

    :param mask: the logical OR of the ``SNAPM_DEBUG_*``
                 values to log.
    :rtype: None
    """
    # pylint: disable=global-statement
    global __DEBUG_MASK
    if mask < 0 or mask > SNAPM_DEBUG_ALL:
        raise ValueError(f"Invalid snapm debug mask: {mask}")
    __DEBUG_MASK = mask


#
# Snapm exception types
#


class SnapmError(Exception):
    """
    Base class for snapshot manager errors.
    """


class SnapmCalloutError(SnapmError):
    """
    An error calling out to an external program.
    """


class SnapmNoSpaceError(SnapmError):
    """
    Insufficient space is available for the requested operation.
    """


class SnapmNoProviderError(SnapmError):
    """
    No snapshot provider plugin was found.
    """


class SnapmSizePolicyError(SnapmError):
    """
    An invalid size policy was specified.
    """


class SnapmExistsError(SnapmError):
    """
    The named snapshot set already exists.
    """


class SnapmBusyError(SnapmError):
    """
    A resouce needed by the current command is already in use: for e.g.
    a snapshot merge is in progress for a previous snapshot set.
    """


class SnapmPathError(SnapmError):
    """
    An invalid path was supplied, for example attempting to snapshot
    something that is not a mount point or block device.
    """


class SnapmNotFoundError(SnapmError):
    """
    The requested object does not exist.
    """


class SnapmInvalidIdentifierError(SnapmError):
    """
    An invalid identifier was given.
    """


class SnapmParseError(SnapmError):
    """
    An error parsing user input.
    """


class SnapmPluginError(SnapmError):
    """
    An error performing an action via a plugin.
    """


class SnapmStateError(SnapmError):
    """
    The status of an object does not allow an operation to proceed.
    """


class SnapmRecursionError(SnapmError):
    """
    A snapshot set source corresponds to a snapshot of another snapshot
    set.
    """


class SnapmArgumentError(SnapmError):
    """
    An invalid argument was passed to a snapshot manager API call.
    """


#
# Selection criteria class
#


# pylint: disable=too-many-instance-attributes
class Selection:
    """
    Selection criteria for snapshot sets and snapshots.
    """

    # Snapshot set fields
    name = None
    uuid = None
    timestamp = None
    nr_snapshots = None
    mount_points = None

    # Snapshot fields
    origin = None
    mount_point = None
    snapshot_name = None
    snapshot_uuid = None

    snapshot_set_attrs = [
        "name",
        "uuid",
        "timestamp",
        "nr_snapshots",
        "mount_points",
    ]

    snapshot_attrs = [
        "origin",
        "mount_point",
        "snapshot_name",
        "snapshot_uuid",
    ]

    all_attrs = snapshot_set_attrs + snapshot_attrs

    def __str__(self):
        """
        Format this ``Selection`` object as a human readable string.

        :returns: A human readable string representation of this
                  Selection object
        :rtype: string
        """
        all_attrs = self.all_attrs
        attrs = [attr for attr in all_attrs if self.__attr_has_value(attr)]
        strval = ""
        tail = ", "
        for attr in set(attrs):
            strval += f"{attr}='{getattr(self, attr)}'{tail}"
        return strval.rstrip(tail)

    def __repr__(self):
        """
        Format this ``Selection`` object as a machine readable string.

        The returned string may be passed to the Selection
        initialiser to duplicate the original Selection.

        :returns: A machine readable string representation of this
                  Selection object
        :rtype: string
        """
        return "Selection(" + str(self) + ")"

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name=None,
        uuid=None,
        timestamp=None,
        nr_snapshots=None,
        mount_points=None,
        origin=None,
        mount_point=None,
        snapshot_name=None,
        snapshot_uuid=None,
    ):
        self.name = name
        self.uuid = uuid
        self.timestamp = timestamp
        self.nr_snapshots = nr_snapshots
        self.mount_points = mount_points
        self.origin = origin
        self.mount_point = mount_point
        self.snapshot_name = snapshot_name
        self.snapshot_uuid = snapshot_uuid

    @classmethod
    def from_cmd_args(cls, cmd_args):
        """
        Initialise Selection from command line arguments.

        Construct a new ``Selection`` object from the command line
        arguments in ``cmd_args``. Each set selection attribute from
        ``cmd_args`` is copied into the Selection. The resulting
        object may be passed to either the ``SnapshotSet`` or
        ``Snapshot`` search functions.
        (``find_snapshot_sets``, ``find_snapshots``, as well as
        the ``snapm.command`` calls that accept a selection argument.

        :param cmd_args: The command line selection arguments.
        :returns: A new Selection instance
        :rtype: Selection
        """
        name = None
        uuid = None
        snapshot_name = None
        snapshot_uuid = None
        if cmd_args.identifier:
            try:
                uuid = UUID(cmd_args.identifier)
            except (TypeError, ValueError):
                name = cmd_args.identifier
        else:
            if cmd_args.name:
                name = cmd_args.name
            elif cmd_args.uuid:
                uuid = cmd_args.uuid

        if hasattr(cmd_args, "snapshot_name"):
            snapshot_name = cmd_args.snapshot_name
        if hasattr(cmd_args, "snapshot_uuid"):
            snapshot_uuid = cmd_args.snapshot_uuid

        select = Selection(
            name=name,
            uuid=uuid,
            snapshot_name=snapshot_name,
            snapshot_uuid=snapshot_uuid,
        )

        _log_debug("Initialised %s from arguments", repr(select))
        return select

    def __attr_has_value(self, attr):
        """
        Test whether an attribute is defined.

        Return ``True`` if the specified attribute name is currently
        defined, or ``False`` otherwise.

        :param attr: The name of the attribute to test
        :returns: ``True`` if ``attr`` is set or ``False`` otherwise
        :rtype: bool
        """
        return hasattr(self, attr) and getattr(self, attr) is not None

    def check_valid_selection(self, snapshot_set=False, snapshot=False):
        """
        Check a Selection for valid criteria.

        Check this ``Selection`` object to ensure it contains only
        criteria that are valid for the specified object type(s).

        Returns ``None`` if the object passes the check, or raise
        ``ValueError`` if invalid criteria exist.

        :param snapshot_set ``Selection`` may include SnapshotSet data
        :param snapshot: ``Selection`` may include Snapshot data
        :returns: ``None`` on success
        :rtype: ``NoneType``
        :raises: ``ValueError`` if excluded criteria are present
        """
        valid_attrs = []
        invalid_attrs = []

        if snapshot_set:
            valid_attrs += self.snapshot_set_attrs
        if snapshot:
            valid_attrs += self.snapshot_attrs

        for attr in self.all_attrs:
            if self.__attr_has_value(attr) and attr not in valid_attrs:
                invalid_attrs.append(attr)

        if invalid_attrs:
            invalid = ", ".join(invalid_attrs)
            raise ValueError(f"Invalid criteria for selection type: {invalid}")

    def is_null(self):
        """
        Test this Selection object for null selection criteria.

        Return ``True`` if this ``Selection`` object matches all
        objects, or ``False`` otherwise.

        :returns: ``True`` if this Selection is null
        :rtype: bool
        """
        all_attrs = self.all_attrs
        attrs = [attr for attr in all_attrs if self.__attr_has_value(attr)]
        return not any(attrs)

    def is_single(self):
        """
        Test this Selection object for single item selction criteria.

        Returns ``True`` if this ``Selection`` object matches a single
        object or ``False`` otherwise.

        A ``Selection`` object matches a single object if either the
        ``name`` or ``uuid`` fields is specified.

        :returns: ``True`` if this selection is single or ``False``
                  otherwise.
        :rtype: bool
        """
        return self.name is not None or self.uuid is not None


def size_fmt(value):
    """
    Format a size in bytes as a human readable string.

    :param value: The integer value to format.
    :returns: A human readable string reflecting value.
    """
    suffixes = ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB"]
    if value == 0:
        return "0B"
    magnitude = int(math.floor(math.log(value, 1024)))
    val = value / math.pow(1024, magnitude)
    if magnitude > 7:
        return f"{val:.1f}YiB"
    return f"{val:3.1f}{suffixes[magnitude]}"


def parse_size_with_units(value):
    """
    Parse a size string with optional unit suffix and return a value in bytes,

    :param size: The size string to parse.
    :returns: an integer size in bytes.
    :raises: ``ValueError`` if the string could not  be parsed as a valid size
             vale.
    """
    match = _SIZE_RE.search(value)
    if match is None:
        raise SnapmSizePolicyError(f"Malformed size expression: {value}")
    (size, unit) = (match.group("size"), match.group("units").upper())
    size_bytes = int(size) * _SIZE_SUFFIXES[unit[0]]
    return size_bytes


def bool_to_yes_no(value):
    """
    Convert boolean-like to yes/no string.

    Return the string "yes" if ``value`` evaluates to ``True``, or "no"
    otherwise.

    :param value: A boolean value to interpet.
    :returns: A string value that is "yes" if ``value`` is ``True``, or "no"
              otherwise.
    """
    return "yes" if bool(value) is True else "no"


def is_size_policy(policy):
    """
    Test whether a string is a valid size policy

    :param policy: A possible policy string to test.
    :returns: ``True`` if ``policy`` is a valid size policy string or
              ``False`` otherwise.
    """
    try:
        _ = SizePolicy("/", "/", 0, 0, 0, policy)
        return True
    except SnapmSizePolicyError:
        return False


class SizePolicyType(Enum):
    """
    Enum class representing the possible snapshot size policies.
    """

    FIXED = "FIXED"
    PERCENT_FREE = "FREE"
    PERCENT_USED = "USED"
    PERCENT_SIZE = "SIZE"


class SizePolicy:
    """
    Class representing a configured instance of a size policy.
    """

    def _parse_size_policy(self, policy):
        """
        Parse a string representation of a size policy.
        """
        if policy is None:
            if self._mount:
                self._percent = DEFAULT_PERCENT_USED
                return SizePolicyType.PERCENT_USED
            self._percent = DEFAULT_PERCENT_SIZE
            return SizePolicyType.PERCENT_SIZE
        if "%" not in policy:
            self._size = parse_size_with_units(policy)
            return SizePolicyType.FIXED
        (percent, policy_type) = policy.rsplit("%", maxsplit=1)
        self._percent = float(percent)
        cap_percent = (SizePolicyType.PERCENT_SIZE, SizePolicyType.PERCENT_FREE)
        for ptype in SizePolicyType:
            if ptype.value == policy_type:
                if ptype == SizePolicyType.PERCENT_USED and not self._mount:
                    raise SnapmSizePolicyError(
                        f"Cannot apply %USED size policy to unmounted block device {self._source}"
                    )
                if ptype in cap_percent and self._percent > 100.0:
                    raise SnapmSizePolicyError(
                        f"Size {self._mount}:{self._percent}%{ptype.value} "
                        "cannot be greater than 100%"
                    )
                return ptype
        raise SnapmSizePolicyError(f"Could not parse size policy: {policy}")

    def __init__(self, source, mount, free_space, fs_used, dev_size, policy):
        """
        Initialise a new `SizePolicy` object using the supplied parameters.

        :param source: The source for this SizePolicy: a mount point or block
                       device path.
        :param mount: The mount point path if mounted.
        :param free_space: The free space available to provision the snapshot
                           in bytes.
        :param fs_used: The current file system occupancy in bytes.
        :param dev_size: The origin device size in bytes.
        :param policy: A size policy string.
        :returns: A `SizePolicy` object configured for the specified size.
        """
        self._source = source
        self._mount = mount
        self._free_space = free_space
        self._fs_used = fs_used
        self._dev_size = dev_size
        self._percent = 0
        self._size = 0
        self.type = self._parse_size_policy(policy)

    @property
    def size(self):
        """
        Return the backing store size for a snapshot based on the policy
        configuration.
        """

        def percent_of_sectors(percent, value):
            size = int(math.ceil((percent * value) / 100))
            return SECTOR_SIZE * math.ceil(size / SECTOR_SIZE)

        if self.type == SizePolicyType.FIXED:
            return self._size
        if self.type == SizePolicyType.PERCENT_FREE:
            return percent_of_sectors(self._percent, self._free_space)
        if self.type == SizePolicyType.PERCENT_USED:
            return percent_of_sectors(self._percent, self._fs_used)
        if self.type == SizePolicyType.PERCENT_SIZE:
            return percent_of_sectors(self._percent, self._dev_size)
        raise SnapmParseError(f"Invalid size policy type: {self.type}")


class SnapStatus(Enum):
    """
    Enum class representing snapshot status: Active, Inactive, Invalid or
    Reverting.
    """

    ACTIVE = 1
    INACTIVE = 2
    INVALID = 3
    REVERTING = 4

    def __str__(self):
        """
        Return a string representation of this ``SnapStatus`` object.

        :returns: "Active", "Inactive", "Invalid", or "Reverting".
        """
        if self == SnapStatus.ACTIVE:
            return "Active"
        if self == SnapStatus.INACTIVE:
            return "Inactive"
        if self == SnapStatus.REVERTING:
            return "Reverting"
        return "Invalid"


# pylint: disable=too-many-public-methods
class SnapshotSet:
    """
    Representation of a set of snapshots taken at the same point
    in time and managed as a group.
    """

    def _link_snapshots(self):
        self._by_mount_point = {}
        self._by_origin = {}
        for snapshot in self._snapshots:
            snapshot.snapshot_set = self
            if snapshot.mount_point:
                self._by_mount_point[snapshot.mount_point] = snapshot
            self._by_origin[snapshot.origin] = snapshot

    def __init__(self, name, timestamp, snapshots):
        """
        Initialise a new ``SnapshotSet`` object.

        :param name: The name of this ``SnapshotSet``.
        :param timestamp: The creation timestamp of this ``SnapshotSet``.
        :param snapshots: A list of ``Snapshot`` objects to be included in this set.
        """
        self._name = name
        self._uuid = uuid5(NAMESPACE_SNAPSHOT_SET, name + str(timestamp))
        self._timestamp = timestamp
        self._snapshots = snapshots
        self._link_snapshots()
        self.boot_entry = None
        self.revert_entry = None

    def __str__(self):
        """
        Return a string representation of this ``SnapshotSet``.

        :returns: A multi-line string describing this snapshot set.
        """
        snapset_str = (
            f"{SNAPSET_NAME}:      {self.name}\n"
            f"{SNAPSET_SOURCES}:          {', '.join(self.sources)}\n"
        )

        if len(self.mount_points) > 0 and len(self.devices) > 0:
            snapset_str += (
                f"  {SNAPSET_MOUNT_POINTS}:    {', '.join(self.mount_points)}\n"
                f"  {SNAPSET_DEVICES}:        {', '.join(self.devices)}\n"
            )

        snapset_str += (
            f"{SNAPSET_NR_SNAPSHOTS}:      {self.nr_snapshots}\n"
            f"{SNAPSET_TIME}:             {datetime.fromtimestamp(self.timestamp)}\n"
            f"{SNAPSET_UUID}:             {self.uuid}\n"
            f"{SNAPSET_STATUS}:           {str(self.status)}\n"
            f"{SNAPSET_AUTOACTIVATE}:     {'yes' if self.autoactivate else 'no'}\n"
            f"{SNAPSET_BOOTABLE}:         {'yes' if self.boot_entry is not None else 'no'}"
        )

        if self.boot_entry or self.revert_entry:
            snapset_str += f"\n{SNAPSET_BOOT_ENTRIES}:"
        if self.boot_entry:
            snapset_str += (
                f"\n  {SNAPSET_SNAPSHOT_ENTRY}:  {self.boot_entry.disp_boot_id}"
            )
        if self.revert_entry:
            snapset_str += (
                f"\n  {SNAPSET_REVERT_ENTRY}:    {self.revert_entry.disp_boot_id}"
            )
        return snapset_str

    def as_dict(self, members=False):
        """
        Return a representation of this snapshot as a dictionary.
        """
        pmap = {}
        pmap[SNAPSET_NAME] = self.name
        pmap[SNAPSET_SOURCES] = self.sources
        pmap[SNAPSET_MOUNT_POINTS] = self.mount_points
        pmap[SNAPSET_DEVICES] = self.devices
        pmap[SNAPSET_NR_SNAPSHOTS] = self.nr_snapshots
        pmap[SNAPSET_TIMESTAMP] = self.timestamp
        pmap[SNAPSET_TIME] = self.time
        pmap[SNAPSET_UUID] = str(self.uuid)
        pmap[SNAPSET_STATUS] = str(self.status)
        pmap[SNAPSET_AUTOACTIVATE] = self.autoactivate
        pmap[SNAPSET_BOOTABLE] = self.boot_entry is not None

        if self.boot_entry or self.revert_entry:
            pmap[SNAPSET_BOOT_ENTRIES] = {}
        if self.boot_entry:
            pmap[SNAPSET_BOOT_ENTRIES][
                SNAPSET_SNAPSHOT_ENTRY
            ] = self.boot_entry.disp_boot_id
        if self.revert_entry:
            pmap[SNAPSET_BOOT_ENTRIES][
                SNAPSET_REVERT_ENTRY
            ] = self.revert_entry.disp_boot_id

        if members:
            pmap[SNAPSET_SNAPSHOTS] = []
            for snapshot in self.snapshots:
                pmap[SNAPSET_SNAPSHOTS].append(snapshot.as_dict())

        return pmap

    def json(self, members=False, pretty=False):
        """
        Return a string representation of this ``SnapshotSet`` in JSON notation.
        """
        return json.dumps(self.as_dict(members=members), indent=4 if pretty else None)

    @property
    def name(self):
        """
        The name of this snapshot set.
        """
        return self._name

    @property
    def uuid(self):
        """
        The UUID of this snapshot set.
        """
        return self._uuid

    @property
    def timestamp(self):
        """
        The numerical timestamp of this snapshot set.
        """
        return self._timestamp

    @property
    def time(self):
        """
        The human readable timestamp of this snapshot set.
        """
        return str(datetime.fromtimestamp(self.timestamp))

    @property
    def snapshots(self):
        """
        The list of snapshots in this snapshot set.
        """
        return self._snapshots

    @property
    def nr_snapshots(self):
        """
        The number of snapshots in this snapshot set.
        """
        return len(self._snapshots)

    @property
    def sources(self):
        """
        The list of souce mount points and block devices in this snapshot set.
        """
        return [s.source for s in self.snapshots]

    @property
    def mount_points(self):
        """
        The list of mount points in this snapshot set.
        """
        return [s.mount_point for s in self.snapshots if s.mount_point]

    @property
    def devices(self):
        """
        The list of block devices in this snapshot set.
        """
        return [s.source for s in self.snapshots if s.source not in self.mount_points]

    @property
    def status(self):
        """
        The overall status of this snapshot set. Returns ``SnapStatus.ACTIVE``
        if all members of the set are valid and active, ``SnapStatus.INACTIVE``
        if any members are inactive, or ``SnapStatus.INVALID`` if any member
        of the set is invalid.
        """
        if any(s.status == SnapStatus.INVALID for s in self.snapshots):
            return SnapStatus.INVALID
        if any(s.status == SnapStatus.REVERTING for s in self.snapshots):
            return SnapStatus.REVERTING
        if any(s.status == SnapStatus.INACTIVE for s in self.snapshots):
            return SnapStatus.INACTIVE
        return SnapStatus.ACTIVE

    @property
    def autoactivate(self):
        """
        The overall autoactivation status of this snapshot set. Returns ``True``
        if all snapshots within the set have autoactivate enabled or ``False``
        otherwise.
        """
        return all(s.autoactivate for s in self.snapshots)

    @autoactivate.setter
    def autoactivate(self, value):
        """
        Set the autoactivation status for all snapshots in this snapshot set.
        """
        for snapshot in self.snapshots:
            try:
                snapshot.autoactivate = value
            except SnapmError as err:
                _log_error(
                    "Failed to set autoactivation for snapshot set member '%s': %s",
                    snapshot.name,
                    err,
                )

    @property
    def origin_mounted(self):
        """
        Test whether the origin volumes for this ``SnapshotSet`` are currently
        mounted and in use.

        :returns: ``True`` if any of the snaphots belonging to this
                  ``SnapshotSet`` are currently mounted, or ``False``
                  otherwise.
        """
        return any(s.origin_mounted for s in self.snapshots)

    @property
    def snapshot_mounted(self):
        """
        Test whether the snapshot volumes for this ``SnapshotSet`` are currently
        mounted and in use.

        :returns: ``True`` if any of the snaphots belonging to this
                  ``SnapshotSet`` are currently mounted, or ``False``
                  otherwise.
        """
        return any(s.snapshot_mounted for s in self.snapshots)

    @property
    def mounted(self):
        """
        Test whether the either the origin or snapshot volumes for this
        ``SnapshotSet`` are currently mounted and in use.

        :returns: ``True`` if any of the snaphots belonging to this
                  ``SnapshotSet`` are currently mounted, or ``False``
                  otherwise.
        """
        return self.snapshot_mounted or self.origin_mounted

    def snapshot_by_mount_point(self, mount_point):
        """
        Return the snapshot corresponding to ``mount_point``.

        :param mount_point: The mount point path to search for.
        :returns: A ``Snapshot`` object for the given mount point.
        :raises: ``SnapmNotFoundError`` if the specified mount point
                 is not present in this ``SnapshotSet``.
        """
        if mount_point in self._by_mount_point:
            return self._by_mount_point[mount_point]
        raise SnapmNotFoundError(
            f"Mount point {mount_point} not found in snapset {self.name}"
        )

    def snapshot_by_source(self, source):
        """
        Return the snapshot corresponding to ``source``.

        :param source: The block device or mount point path to search for.
        :returns: A ``Snapshot`` object for the given source path.
        :raises: ``SnapmNotFoundError`` if the specified source path
                 is not present in this ``SnapshotSet``.
        """
        if source in self._by_mount_point:
            return self._by_mount_point[source]
        if source in self._by_origin:
            return self._by_origin[source]
        raise SnapmNotFoundError(
            f"Source path {source} not found in snapset {self.name}"
        )

    def rename(self, new_name):
        """
        Rename this ``SnapshotSet`` to ``new_name``.

        :raises: ``SnapmError`` if a call to rename any member of the snapshot
                 set fails.
        """
        _log_info(
            "Attempting rename for snapshot set '%s' to '%s'", self.name, new_name
        )
        old_name = self.name
        snapshots = self.snapshots
        new_snapshots = []

        for snapshot in snapshots.copy():
            snapshots.remove(snapshot)
            try:
                _log_debug("Renaming snapshot set member '%s'", snapshot.name)
                new_snapshot = snapshot.rename(new_name)
                new_snapshots.append(new_snapshot)
            except SnapmError as err:
                _log_error("Failed to rename snapshot %s: %s", snapshot.name, err)
                rollback_snapshots = []
                for new_snapshot in new_snapshots:
                    try:
                        rollback_snapshot = snapshot.rename(old_name)
                        rollback_snapshots.append(rollback_snapshot)
                    except SnapmError as err2:
                        _log_error(
                            "Failed to rollback snapshot rename on %s: %s",
                            snapshot.name,
                            err2,
                        )
                self._snapshots = snapshots + rollback_snapshots
                raise SnapmPluginError(
                    f"Could not rename all snapshots for set {old_name}"
                ) from err

        self._name = new_name
        self._uuid = uuid5(NAMESPACE_SNAPSHOT_SET, self.name + str(self.timestamp))
        self._snapshots = new_snapshots
        self._link_snapshots()

    def delete(self):
        """
        Delete the on-disk representation of this ``SnapshotSet``.

        :raises: ``SnapmBusyError`` if the snapshot set is in use.
                 ``SnapmPluginError`` if an error occurs deleting snapshot
                 set members.
        """
        _log_info("Attempting delete for snapshot set '%s'", self.name)
        if any(snapshot.snapshot_mounted for snapshot in self.snapshots):
            raise SnapmBusyError(
                f"Snapshots from snapshot set {self.name} are mounted: cannot delete"
            )
        if self.status == SnapStatus.REVERTING:
            _log_error("Cannot operate on reverting snapshot set '%s'", self.name)
            raise SnapmBusyError(f"Failed to delete snapshot set {self.name}")
        for snapshot in self.snapshots.copy():
            _log_debug("Deleting snapshot set member '%s'", snapshot.name)
            try:
                snapshot.delete()
                self.snapshots.remove(snapshot)
                if snapshot.mount_point:
                    self._by_mount_point.pop(snapshot.mount_point)
                self._by_origin.pop(snapshot.origin)
            except SnapmError as err:
                _log_error(
                    "Failed to delete snapshot set member '%s': %s",
                    snapshot.name,
                    err,
                )
                raise SnapmPluginError(
                    f"Could not delete all snapshots for set {self.name}"
                ) from err

    def resize(self, sources, size_policies):
        """
        Attempt to resize the ``SnapshotSet`` according to the list of
        ``sources`` and corresponding ``size_policy`` strings.

        :raises: ``SnapmNoSpaceError`` if there is insufficient space available
                 for the requested operation.
        """
        _log_info("Attempting resize for snapshot set '%s'", self.name)
        for source in sources:
            try:
                _ = self.snapshot_by_source(source)
            except SnapmNotFoundError as err:
                _log_error(
                    "Cannot resize %s: source path not a member of snapset %s",
                    source,
                    self.name,
                )
                raise err

        providers = set(snapshot.provider for snapshot in self.snapshots)
        for provider in providers:
            provider.start_transaction()

        for source in sources:
            snapshot = self.snapshot_by_source(source)
            size_policy = size_policies[source]
            try:
                snapshot.check_resize(size_policy)
            except SnapmNoSpaceError as err:
                _log_error("Cannot resize %s snapshot: %s", snapshot.name, err)
                raise SnapmNoSpaceError(
                    f"Insufficient free space to resize snapshot set {self.name}"
                ) from err

        for source in sources:
            snapshot = self.snapshot_by_source(source)
            size_policy = size_policies[source]
            _log_debug("Resizing snapshot set member '%s'", snapshot.name)
            try:
                snapshot.resize(size_policy)
            except SnapmNoSpaceError as err:
                _log_error("Cannot resize %s snapshot: %s", snapshot.name, err)
                raise SnapmNoSpaceError(
                    f"Insufficient free space to resize snapshot set {self.name}"
                ) from err

        for provider in providers:
            provider.end_transaction()

    def revert(self):
        """
        Initiate a revert operation on this ``SnapshotSet``.

        :raises: ``SnapmPluginError`` if a plugin fails to perform the
                 requested operation.
        """
        _log_info("Attempting revert for snapshot set '%s'", self.name)
        revert_entry = self.revert_entry
        mounted = self.mounted
        name = self.name

        # Perform revert operation on all snapshots
        for snapshot in self.snapshots:
            try:
                _log_debug("Reverting snapshot set member '%s'", snapshot.name)
                snapshot.revert()
            except SnapmError as err:
                _log_error(
                    "Failed to revert snapshot set member '%s': %s",
                    snapshot.name,
                    err,
                )
                raise SnapmPluginError(
                    f"Could not revert all snapshots for set {self.name}"
                ) from err
        if mounted:
            _log_warn(
                "Snaphot set %s is in use: reboot required to complete revert",
                name,
            )
            if revert_entry:
                _log_warn(
                    "Boot into '%s' to continue",
                    revert_entry.title,
                )

    def activate(self):
        """
        Attempt to activate all members of this ``SnapshotSet``.

        :raises: ``SnapmPluginError`` if a plugin fails to perform the
                 requested operation.
        """
        _log_info("Attempting to activate snapshot set '%s'", self.name)
        for snapshot in self.snapshots:
            try:
                _log_debug("Activating snapshot set member '%s'", snapshot.name)
                snapshot.activate()
            except SnapmError as err:
                _log_error(
                    "Failed to activate snapshot set member '%s': %s",
                    snapshot.name,
                    err,
                )
                raise SnapmPluginError(
                    f"Could not activate all snapshots for set {self.name}"
                ) from err

    def deactivate(self):
        """
        Attempt to deactivate all members of this ``SnapshotSet``.

        :raises: ``SnapmPluginError`` if a plugin fails to perform the
                 requested operation.
        """
        _log_info("Attempting to deactivate snapshot set '%s'", self.name)
        for snapshot in self.snapshots:
            try:
                _log_debug("Deactivating snapshot set member '%s'", snapshot.name)
                snapshot.deactivate()
            except SnapmError as err:
                _log_error(
                    "Failed to deactivate snapshot set member '%s': %s",
                    snapshot.name,
                    err,
                )
                raise SnapmPluginError(
                    f"Could not deactivate all snapshots for set {self.name}"
                ) from err


# pylint: disable=too-many-instance-attributes,too-many-public-methods
class Snapshot:
    """
    Base class for individual snapshots. Each snapshot plugin should
    subclass ``Snapshot`` to provide a specific implementation.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, name, snapset_name, origin, timestamp, mount_point, provider):
        """
        Initialise a new ``Snapshot`` object.

        :param name: The name of the snapshot.
        :param snapset_name: The name of the ``SnapshotSet`` this snapshot is a part of.
        :param origin: The origin volume of the snapshot.
        :param timestamp: The creation timestamp of the snapshot set.
        :param mount_point: The mount point path this snapshot refers to.
        :param provider: The plugin providing this snapshot.
        """
        self._name = name
        self._uuid = uuid5(NAMESPACE_SNAPSHOT, name)
        self._snapset_name = snapset_name
        self._origin = origin
        self._timestamp = timestamp
        self._mount_point = mount_point
        self._snapshot_set = None
        self.provider = provider

    def __str__(self):
        """
        Return a string representation of this ``Snapshot`` object.

        :returns: A multi-line string describing this snapshot.
        """
        return (
            f"{SNAPSHOT_NAME}:           {self.name}\n"
            f"{SNAPSET_NAME}:    {self.snapset_name}\n"
            f"{SNAPSHOT_ORIGIN}:         {self.origin}\n"
            f"{SNAPSET_TIME}:           {datetime.fromtimestamp(self.timestamp)}\n"
            f"{SNAPSHOT_SOURCE}:         {self.source}\n"
            f"{SNAPSHOT_MOUNT_POINT}:     {self.mount_point}\n"
            f"{SNAPSHOT_PROVIDER}:       {self.provider.name}\n"
            f"{SNAPSHOT_UUID}:           {self.uuid}\n"
            f"{SNAPSHOT_STATUS}:         {str(self.status)}\n"
            f"{SNAPSHOT_SIZE}:           {size_fmt(self.size)}\n"
            f"{SNAPSHOT_FREE}:           {size_fmt(self.free)}\n"
            f"{SNAPSHOT_AUTOACTIVATE}:   {'yes' if self.autoactivate else 'no'}\n"
            f"{SNAPSHOT_DEV_PATH}:     {self.devpath}"
        )

    def as_dict(self):
        """
        Return a representation of this snapshot as a dictionary.
        """
        pmap = {}
        pmap[SNAPSHOT_NAME] = self.name
        pmap[SNAPSET_NAME] = self.snapset_name
        pmap[SNAPSHOT_ORIGIN] = self.origin
        pmap[SNAPSET_TIMESTAMP] = self.timestamp
        pmap[SNAPSET_TIME] = self.time
        pmap[SNAPSHOT_SOURCE] = self.source
        pmap[SNAPSHOT_MOUNT_POINT] = self.mount_point
        pmap[SNAPSHOT_PROVIDER] = self.provider.name
        pmap[SNAPSHOT_UUID] = str(self.uuid)
        pmap[SNAPSHOT_STATUS] = str(self.status)
        pmap[SNAPSHOT_SIZE] = size_fmt(self.size)
        pmap[SNAPSHOT_FREE] = size_fmt(self.free)
        pmap[SNAPSHOT_SIZE_BYTES] = self.size
        pmap[SNAPSHOT_FREE_BYTES] = self.free
        pmap[SNAPSHOT_AUTOACTIVATE] = self.autoactivate
        pmap[SNAPSHOT_DEV_PATH] = self.devpath
        return pmap

    def json(self, pretty=False):
        """
        Return a string representation of this ``Snapshot`` in JSON notation.
        """
        return json.dumps(self.as_dict(), indent=4 if pretty else None)

    @property
    def name(self):
        """
        The name of this snapshot.
        """
        return self._name

    @property
    def uuid(self):
        """
        The UUID of this snapshot.
        """
        return self._uuid

    @property
    def snapset_name(self):
        """
        The name of the snapshot set this snapshot belongs to.
        """
        return self._snapset_name

    @property
    def origin(self):
        """
        The origin of this snapshot.
        """
        raise NotImplementedError

    @property
    def origin_options(self):
        """
        File system options needed to specify the origin of this snapshot.
        """
        raise NotImplementedError

    @property
    def timestamp(self):
        """
        The numerical timestamp of this snapshot.
        """
        return self._timestamp

    @property
    def time(self):
        """
        The human readable timestamp of this snapshot.
        """
        return str(datetime.fromtimestamp(self.timestamp))

    @property
    def mount_point(self):
        """
        The mount point of this snapshot.
        """
        return self._mount_point

    @property
    def source(self):
        """
        The block device or mount point from which this ``Snapshot``
        was created.
        """
        return self.mount_point if self.mount_point else self.origin

    @property
    def snapshot_set(self):
        """
        The ``SnapshotSet`` this snapshot belongs to.
        """
        return self._snapshot_set

    @snapshot_set.setter
    def snapshot_set(self, value):
        """
        Set the ``SnapshotSet`` this snapshot belongs to.
        """
        self._snapshot_set = value

    @property
    def devpath(self):
        """
        The device path for this snapshot.
        """
        raise NotImplementedError

    @property
    def status(self):
        """
        The status of this snapshot. Returns a ``SnapStatus`` enum
        value representing the current state of the snapshot.
        """
        raise NotImplementedError

    @property
    def size(self):
        """
        The size of this snapshot in bytes. For snapshots with fixed-size
        backstores this reflects the size of the backstore. For snapshots
        that dynamically allocate space to the snapshot it reflects the
        device or volume size.
        """
        raise NotImplementedError

    @property
    def free(self):
        """
        The space available to this snapshot in bytes. For snapshots with
        fixed-size backstores this reflects the space remaining in the
        current backstore. For snapshots that dynamically allocate space
        to the snapshot it indicates the pooled space available.
        """
        raise NotImplementedError

    @property
    def autoactivate(self):
        """
        The autoactivation status of this snapshot. Returns ``True`` if the
        snapshot is automatically activated or ``False`` otherwise.
        """
        raise NotImplementedError

    @autoactivate.setter
    def autoactivate(self, value):
        """
        Set the autoactivation state of this snapshot.

        :param value: ``True`` to enable autoactivation or ``False`` otherwise.
        """
        self.provider.set_autoactivate(self.name, auto=value)
        self.invalidate_cache()

    @property
    def origin_mounted(self):
        """
        Test whether the origin volume for this ``Snapshot`` is currently
        mounted and in use.

        :returns: ``True`` if this snaphot's prigin is currently mounted
                  or ``False`` otherwise.
        """
        with open("/proc/mounts", "r", encoding="utf8") as mounts:
            for line in mounts:
                fields = line.split(" ")
                if self.mount_point == fields[1]:
                    return os.path.samefile(self.origin, fields[0])
        return False

    @property
    def snapshot_mounted(self):
        """
        Test whether the snapshot volume for this ``Snapshot`` is currently
        mounted and in use.

        :returns: ``True`` if this snaphot is currently mounted or ``False``
                  otherwise.
        """
        if self.status != SnapStatus.ACTIVE:
            return False
        with open("/proc/mounts", "r", encoding="utf8") as mounts:
            for line in mounts:
                fields = line.split(" ")
                devpath = fields[0]
                if os.path.exists(devpath):
                    if os.path.samefile(self.devpath, devpath):
                        return True
        return False

    @property
    def mounted(self):
        """
        Test whether either the snapshot or origin volume for this ``Snapshot``
        is currently mounted and in use.

        :returns: ``True`` if this snapshot or its origin is currently mounted
                  or ``False`` otherwise.
        """
        return self.snapshot_mounted or self.origin_mounted

    def delete(self):
        """
        Delete this snapshot.
        """
        self.provider.delete_snapshot(self.name)

    def rename(self, new_name):
        """
        Rename a snapshot within a snapshot set.

        :param new_name: The new name for the snapshot.
        """
        return self.provider.rename_snapshot(
            self.name, self.origin, new_name, self.timestamp, self.mount_point
        )

    def check_resize(self, size_policy):
        """
        Check whether this snapshot can be resized or not by calling the
        provider plugin with the updated ``size_policy``.
        """
        self.invalidate_cache()
        self.provider.check_resize_snapshot(
            self.name, self.origin, self.mount_point, size_policy
        )

    def resize(self, size_policy):
        """
        Attempt to resize this snapshot according to the updated
        ``size_policy``.
        """
        self.invalidate_cache()
        self.provider.resize_snapshot(
            self.name, self.origin, self.mount_point, size_policy
        )

    def check_revert(self):
        """
        Check whether this snapshot can be reverted or not by calling the
        provider plugin to verify the state of the snapshot.
        """
        self.provider.check_revert_snapshot(self.name, self.origin)

    def revert(self):
        """
        Request to revert a snapshot and revert the content of the origin
        volume to its state at the time of the snapshot.

        This may be deferred until the next device activation or mount
        operation for the respective volume.
        """
        self.invalidate_cache()
        self.provider.revert_snapshot(self.name)

    def invalidate_cache(self):
        """
        Invalidate any cached data describing this snapshot.
        """
        raise NotImplementedError

    def activate(self):
        """
        Activate this snapshot.
        """
        self.provider.activate_snapshot(self.name)
        self.invalidate_cache()

    def deactivate(self):
        """
        Deactivate this snapshot.
        """
        self.provider.deactivate_snapshot(self.name)
        self.invalidate_cache()


__all__ = [
    "ETC_FSTAB",
    "SNAPSET_NAME",
    "SNAPSET_SOURCES",
    "SNAPSET_MOUNT_POINTS",
    "SNAPSET_DEVICES",
    "SNAPSET_NR_SNAPSHOTS",
    "SNAPSET_TIME",
    "SNAPSET_TIMESTAMP",
    "SNAPSET_UUID",
    "SNAPSET_STATUS",
    "SNAPSET_AUTOACTIVATE",
    "SNAPSET_BOOTABLE",
    "SNAPSET_BOOT_ENTRIES",
    "SNAPSET_SNAPSHOT_ENTRY",
    "SNAPSET_REVERT_ENTRY",
    "SNAPSET_SNAPSHOTS",
    "SNAPSHOT_NAME",
    "SNAPSHOT_ORIGIN",
    "SNAPSHOT_SOURCE",
    "SNAPSHOT_MOUNT_POINT",
    "SNAPSHOT_PROVIDER",
    "SNAPSHOT_UUID",
    "SNAPSHOT_STATUS",
    "SNAPSHOT_SIZE",
    "SNAPSHOT_FREE",
    "SNAPSHOT_SIZE_BYTES",
    "SNAPSHOT_FREE_BYTES",
    "SNAPSHOT_AUTOACTIVATE",
    "SNAPSHOT_DEV_PATH",
    "SNAPM_VALID_NAME_CHARS",
    "SNAPM_DEBUG_MANAGER",
    "SNAPM_DEBUG_COMMAND",
    "SNAPM_DEBUG_REPORT",
    "SNAPM_DEBUG_ALL",
    "SnapmLogger",
    "set_debug_mask",
    "get_debug_mask",
    "SnapmError",
    "SnapmCalloutError",
    "SnapmNoSpaceError",
    "SnapmNoProviderError",
    "SnapmSizePolicyError",
    "SnapmExistsError",
    "SnapmBusyError",
    "SnapmPathError",
    "SnapmNotFoundError",
    "SnapmInvalidIdentifierError",
    "SnapmParseError",
    "SnapmPluginError",
    "SnapmStateError",
    "SnapmRecursionError",
    "SnapmArgumentError",
    "Selection",
    "size_fmt",
    "is_size_policy",
    "parse_size_with_units",
    "bool_to_yes_no",
    "SizePolicyType",
    "SizePolicy",
    "SnapStatus",
    "SnapshotSet",
    "Snapshot",
]



# Copyright Red Hat
#
# tests/__init__.py - Snapshot Manager test package
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
import logging
import os
from os.path import abspath, join
import time

log = logging.getLogger()
log.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s %(levelname)s %(name)s %(message)s')
file_handler = logging.FileHandler("test.log")
file_handler.setFormatter(formatter)
console_handler = logging.StreamHandler()
console_handler.setFormatter(formatter)
log.addHandler(file_handler)
log.addHandler(console_handler)

os.environ["TZ"] = "UTC"
time.tzset()

BOOT_ROOT_TEST = join(os.getcwd(), "tests/boot")


class MockArgs(object):
    identifier = None
    debug = None
    name = None
    new_name = None
    name_prefixes = False
    no_headings = False
    options = ""
    members = False
    sort = ""
    rows = False
    separator = None
    uuid = None
    verbose = 0
    version = False
    json = False


def have_root():
    """Return ``True`` if the test suite is running as the root user,
    and ``False`` otherwise.
    """
    return os.geteuid() == 0 and os.getegid() == 0



# Copyright Red Hat
#
# snapm/manager/signals.py - Snapshot Manager signal handling
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
"""
Helpers for blocking and unblocking signal delivery.
"""
from signal import SIG_BLOCK, SIG_UNBLOCK, SIGINT, SIGTERM, pthread_sigmask
from functools import wraps
import logging

_log = logging.getLogger(__name__)
_log_debug = _log.debug

_to_block = {SIGINT, SIGTERM}


def block_signals():
    """
    Block termination signals before entering critical section.
    """
    _log_debug("Blocking termination signals %s", _to_block)
    pthread_sigmask(SIG_BLOCK, _to_block)


def unblock_signals():
    """
    Unblock and deliver termination signals after leaving critical section.
    """
    _log_debug("Unblocking termination signals %s", _to_block)
    pthread_sigmask(SIG_UNBLOCK, _to_block)


def suspend_signals(func):
    """
    Decorator to wrap functions that implement a critical section.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        block_signals()
        try:
            ret = func(*args, **kwargs)
        finally:
            unblock_signals()
        return ret

    return wrapper



# Copyright Red Hat
#
# tests/test_report.py - report API tests.
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
import unittest
import logging
import uuid
from io import StringIO

log = logging.getLogger()

from snapm.report import (
    REP_NUM,
    REP_STR,
    REP_SHA,
    REP_TIME,
    REP_UUID,
    REP_SIZE,
    REP_STR_LIST,
    ALIGN_LEFT,
    ALIGN_RIGHT,
    ReportOpts,
    ReportObjType,
    FieldType,
    Report,
)

_report_objs = [
    (
        1,
        "foo",
        "ffffffffffffffffffffffffffffffffffffffff",
        "2023-09-05 14:40:53",
        uuid.UUID("00000000-0000-0000-0000-000000000000"),
        1024,
        ["foo", "bar", "baz"],
    ),
    (
        2,
        "bar",
        "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF",
        "1978-01-10 09:13:12",
        uuid.UUID("1d133bcc-6137-5267-b870-e469f7188dbe"),
        1024**2,
        ["one", "two", "three"],
    ),
    (
        3,
        "baz",
        "1111111111111111111111111111111111111111",
        "2022-08-01 16:43:32",
        uuid.UUID("3576355e-e4d7-5a57-9f24-b3f4e0e326ef"),
        1024**3,
        ["baz", "quux"],
    ),
    (
        4,
        "qux",
        "2222222222222222222222222222222222222222",
        "2020-01-01 23:32:08",
        uuid.UUID("046e4f15-1ddd-5724-b032-878248a71d4b"),
        1024**4,
        ["string", "list"],
    ),
]

PR_NUM = 1
PR_STR = 2
PR_SHA = 4
PR_TIME = 8
PR_UUID = 16
PR_SIZE = 32
PR_STR_LIST = 64

_test_obj_types = [
    ReportObjType(PR_NUM, "Num", "num_", lambda o: o[0]),
    ReportObjType(PR_STR, "Str", "str_", lambda o: o[1]),
    ReportObjType(PR_SHA, "Sha", "sha_", lambda o: o[2]),
    ReportObjType(PR_TIME, "Time", "time_", lambda o: o[3]),
    ReportObjType(PR_UUID, "Uuid", "uuid_", lambda o: o[4]),
    ReportObjType(PR_SIZE, "Size", "size_", lambda o: o[5]),
    ReportObjType(PR_STR_LIST, "StrList", "strlist_", lambda o: o[6]),
]


class ReportTests(unittest.TestCase):
    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)

    def tearDown(self):
        log.debug("Tearing down (%s)", self._testMethodName)

    def test_FieldType_no_type(self):
        with self.assertRaises(ValueError):
            FieldType(0, None, "None", "Nothing", 0, REP_NUM, lambda x: x)

    def test_FieldType_no_name(self):
        with self.assertRaises(ValueError):
            FieldType(PR_NUM, None, "None", "Nothing", 0, REP_NUM, lambda x: x)

    def test_FieldType_bogus_dtype_raises(self):
        with self.assertRaises(ValueError):
            FieldType(PR_NUM, "none", "None", "Nothing", 0, "fzzrt", lambda x: x)

    def test_FieldType_dtype_NUM(self):
        pf = FieldType(PR_NUM, "none", "None", "Nothing", 0, REP_NUM, lambda x: x)
        self.assertEqual(pf.dtype, REP_NUM)

    def test_FieldType_dtype_STR(self):
        pf = FieldType(PR_STR, "none", "None", "Nothing", 0, REP_STR, lambda x: x)
        self.assertEqual(pf.dtype, REP_STR)

    def test_FieldType_dtype_SHA(self):
        pf = FieldType(PR_SHA, "none", "None", "Nothing", 0, REP_SHA, lambda x: x)
        self.assertEqual(pf.dtype, REP_SHA)

    def test_FieldType_dtype_TIME(self):
        pf = FieldType(PR_TIME, "none", "None", "Nothing", 0, REP_TIME, lambda x: x)
        self.assertEqual(pf.dtype, REP_TIME)

    def test_FieldType_dtype_UUID(self):
        pf = FieldType(PR_UUID, "none", "None", "Nothing", 0, REP_UUID, lambda x: x)
        self.assertEqual(pf.dtype, REP_UUID)

    def test_FieldType_dtype_SIZE(self):
        pf = FieldType(PR_SIZE, "none", "None", "Nothing", 0, REP_SIZE, lambda x: x)
        self.assertEqual(pf.dtype, REP_SIZE)

    def test_FieldType_dtype_STR_LIST(self):
        pf = FieldType(
            PR_STR_LIST, "none", "non", "Nothing", 0, REP_STR_LIST, lambda x: x
        )
        self.assertEqual(pf.dtype, REP_STR_LIST)

    def test_FieldType_bogus_align_raises(self):
        with self.assertRaises(ValueError):
            FieldType(
                PR_NUM,
                "none",
                "None",
                "Nothing",
                0,
                REP_NUM,
                lambda x: x,
                align="qux",
            )

    def test_FieldType_with_align_l(self):
        FieldType(
            PR_NUM,
            "none",
            "None",
            "Nothing",
            0,
            REP_NUM,
            lambda x: x,
            align=ALIGN_LEFT,
        )

    def test_FieldType_with_align_r(self):
        FieldType(
            PR_NUM,
            "none",
            "None",
            "Nothing",
            0,
            REP_NUM,
            lambda x: x,
            align=ALIGN_RIGHT,
        )

    def test_FieldType_negative_width_raises(self):
        with self.assertRaises(ValueError) as cm:
            FieldType(PR_NUM, "none", "None", "Nothing", -1, REP_NUM, lambda x: x)

    def test_FieldType_simple_str_int_report(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )

        output = StringIO()
        opts = ReportOpts(report_file=output)

        xoutput = (
            "Name     Number\n"
            "foo             1\n"
            "bar             2\n"
            "baz             3\n"
            "qux             4\n"
        )

        pr = Report(
            _test_obj_types, [pf_name, pf_num], "name,,number", opts, None, None
        )

        for obj in _report_objs:
            pr.report_object(obj)
        pr.report_output()

        self.assertEqual(output.getvalue(), xoutput)

    def test_FieldType_simple_str_int_report_as_rows(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )

        output = StringIO()
        opts = ReportOpts(columns_as_rows=True, report_file=output)

        xoutput = (
            "Name foo      bar      baz      qux\n"
            "Number        1        2        3        4\n"
        )

        pr = Report(
            _test_obj_types, [pf_name, pf_num], "name,,number", opts, None, None
        )

        for obj in _report_objs:
            pr.report_object(obj)
        pr.report_output()

        self.assertEqual(output.getvalue(), xoutput)

    def test_FieldType_simple_str_int_report_noheadings(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )

        output = StringIO()
        opts = ReportOpts(headings=False, report_file=output)

        xoutput = (
            "foo             1\n"
            "bar             2\n"
            "baz             3\n"
            "qux             4\n"
        )

        pr = Report(
            _test_obj_types, [pf_name, pf_num], "name,,number", opts, None, None
        )

        for obj in _report_objs:
            pr.report_object(obj)
        pr.report_output()

        self.assertEqual(output.getvalue(), xoutput)

    def test_FieldType_simple_str_int_report_default_fields(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )

        output = StringIO()
        opts = ReportOpts(report_file=output)

        xoutput = (
            "Name     Number\n"
            "foo             1\n"
            "bar             2\n"
            "baz             3\n"
            "qux             4\n"
        )

        pr = Report(_test_obj_types, [pf_name, pf_num], None, opts, None, None)

        for obj in _report_objs:
            pr.report_object(obj)
        pr.report_output()

        self.assertEqual(output.getvalue(), xoutput)

    def test_FieldType_simple_str_int_report_with_sort(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )

        output = StringIO()
        opts = ReportOpts(report_file=output)

        xoutput = (
            "Name     Number\n"
            "bar             2\n"
            "baz             3\n"
            "foo             1\n"
            "qux             4\n"
        )

        pr = Report(
            _test_obj_types, [pf_name, pf_num], "name,,number", opts, "name", None
        )

        for obj in _report_objs:
            pr.report_object(obj)
        pr.report_output()

        self.assertEqual(output.getvalue(), xoutput)

    def test_FieldType_simple_str_int_report_with_sort_num(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )

        output = StringIO()
        opts = ReportOpts(report_file=output)

        xoutput = (
            "Name     Number\n"
            "foo             1\n"
            "bar             2\n"
            "baz             3\n"
            "qux             4\n"
        )

        pr = Report(
            _test_obj_types, [pf_name, pf_num], "name,,number", opts, "number", None
        )

        for obj in _report_objs:
            pr.report_object(obj)
        pr.report_output()

        self.assertEqual(output.getvalue(), xoutput)

    def test_FieldType_simple_str_int_report_with_sort_ascend(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )

        output = StringIO()
        opts = ReportOpts(report_file=output)

        xoutput = (
            "Name     Number\n"
            "bar             2\n"
            "baz             3\n"
            "foo             1\n"
            "qux             4\n"
        )

        pr = Report(
            _test_obj_types, [pf_name, pf_num], "name,,number", opts, "+name", None
        )

        for obj in _report_objs:
            pr.report_object(obj)
        pr.report_output()

        self.assertEqual(output.getvalue(), xoutput)

    def test_FieldType_simple_str_int_report_with_sort_descend(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )

        output = StringIO()
        opts = ReportOpts(report_file=output)

        xoutput = (
            "Name     Number\n"
            "qux             4\n"
            "foo             1\n"
            "baz             3\n"
            "bar             2\n"
        )

        pr = Report(
            _test_obj_types, [pf_name, pf_num], "name,,number", opts, "-name", None
        )

        for obj in _report_objs:
            pr.report_object(obj)
        pr.report_output()

        self.assertEqual(output.getvalue(), xoutput)

    def test_FieldType_simple_str_int_report_with_sort_two_fields(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )

        output = StringIO()
        opts = ReportOpts(report_file=output)

        xoutput = (
            "Name     Number\n"
            "bar             2\n"
            "baz             3\n"
            "foo             1\n"
            "qux             4\n"
        )

        pr = Report(
            _test_obj_types,
            [pf_name, pf_num],
            "name,,number",
            opts,
            "name,,number",
            None,
        )

        for obj in _report_objs:
            pr.report_object(obj)
        pr.report_output()

        self.assertEqual(output.getvalue(), xoutput)

    def test_FieldType_simple_str_int_report_with_bad_sort(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )

        output = StringIO()
        opts = ReportOpts(report_file=output)

        with self.assertRaises(ValueError) as cm:
            Report(
                _test_obj_types, [pf_name, pf_num], "name,,number", opts, "nosuch", None
            )

    def test_FieldType_all_types_report(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )
        pf_sha = FieldType(
            PR_SHA,
            "sha",
            "Sha",
            "Nothing",
            8,
            REP_SHA,
            lambda f, d: f.report_sha(d),
        )
        pf_time = FieldType(
            PR_TIME,
            "time",
            "Time",
            "Nothing",
            8,
            REP_TIME,
            lambda f, d: f.report_time(d),
        )
        pf_uuid = FieldType(
            PR_UUID,
            "uuid",
            "Uuid",
            "Nothing",
            8,
            REP_UUID,
            lambda f, d: f.report_uuid(d),
        )
        pf_size = FieldType(
            PR_SIZE,
            "size",
            "Size",
            "Nothing",
            8,
            REP_SIZE,
            lambda f, d: f.report_size(d),
        )
        pf_str_list = FieldType(
            PR_STR_LIST,
            "strlist",
            "StrList",
            "Nothing",
            8,
            REP_STR_LIST,
            lambda f, d: f.report_str_list(d),
        )

        output = StringIO()
        opts = ReportOpts(report_file=output)

        xoutput = (
            "Name     Number   Sha      Time                Uuid                                 Size     StrList\n"
            "foo             1 ffffffff 2023-09-05 14:40:53 00000000-0000-0000-0000-000000000000   1.0KiB bar, baz, foo\n"
            "bar             2 FFFFFFFF 1978-01-10 09:13:12 1d133bcc-6137-5267-b870-e469f7188dbe   1.0MiB one, three, two\n"
            "baz             3 11111111 2022-08-01 16:43:32 3576355e-e4d7-5a57-9f24-b3f4e0e326ef   1.0GiB baz, quux\n"
            "qux             4 22222222 2020-01-01 23:32:08 046e4f15-1ddd-5724-b032-878248a71d4b   1.0TiB list, string\n"
        )

        pr = Report(
            _test_obj_types,
            [pf_name, pf_num, pf_sha, pf_time, pf_uuid, pf_size, pf_str_list],
            "name,number,sha,time,uuid,size,strlist",
            opts,
            None,
            None,
        )

        for obj in _report_objs:
            pr.report_object(obj)
        pr.report_output()
        print(output.getvalue())
        self.assertEqual(output.getvalue(), xoutput)

    def test_FieldType_all_types_report_json(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )
        pf_sha = FieldType(
            PR_SHA,
            "sha",
            "Sha",
            "Nothing",
            8,
            REP_SHA,
            lambda f, d: f.report_sha(d),
        )
        pf_time = FieldType(
            PR_TIME,
            "time",
            "Time",
            "Nothing",
            8,
            REP_TIME,
            lambda f, d: f.report_time(d),
        )
        pf_uuid = FieldType(
            PR_UUID,
            "uuid",
            "Uuid",
            "Nothing",
            8,
            REP_UUID,
            lambda f, d: f.report_uuid(d),
        )
        pf_size = FieldType(
            PR_SIZE,
            "size",
            "Size",
            "Nothing",
            8,
            REP_SIZE,
            lambda f, d: f.report_size(d),
        )
        pf_str_list = FieldType(
            PR_STR_LIST,
            "strlist",
            "StrList",
            "Nothing",
            8,
            REP_STR_LIST,
            lambda f, d: f.report_str_list(d),
        )

        output = StringIO()
        opts = ReportOpts(json=True, report_file=output)

        xoutput = (
            '{\n'
            '    "Test": [\n'
            '        {\n'
            '            "str_name": "foo",\n'
            '            "num_number": 1,\n'
            '            "sha_sha": "ffffffffffffffffffffffffffffffffffffffff",\n'
            '            "time_time": "2023-09-05 14:40:53",\n'
            '            "uuid_uuid": "00000000-0000-0000-0000-000000000000",\n'
            '            "size_size": "1.0KiB",\n'
            '            "strlist_strlist": [\n'
            '                "bar",\n'
            '                "baz",\n'
            '                "foo"\n'
            '            ]\n'
            '        },\n'
            '        {\n'
            '            "str_name": "bar",\n'
            '            "num_number": 2,\n'
            '            "sha_sha": "FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF",\n'
            '            "time_time": "1978-01-10 09:13:12",\n'
            '            "uuid_uuid": "1d133bcc-6137-5267-b870-e469f7188dbe",\n'
            '            "size_size": "1.0MiB",\n'
            '            "strlist_strlist": [\n'
            '                "one",\n'
            '                "three",\n'
            '                "two"\n'
            '            ]\n'
            '        },\n'
            '        {\n'
            '            "str_name": "baz",\n'
            '            "num_number": 3,\n'
            '            "sha_sha": "1111111111111111111111111111111111111111",\n'
            '            "time_time": "2022-08-01 16:43:32",\n'
            '            "uuid_uuid": "3576355e-e4d7-5a57-9f24-b3f4e0e326ef",\n'
            '            "size_size": "1.0GiB",\n'
            '            "strlist_strlist": [\n'
            '                "baz",\n'
            '                "quux"\n'
            '            ]\n'
            '        },\n'
            '        {\n'
            '            "str_name": "qux",\n'
            '            "num_number": 4,\n'
            '            "sha_sha": "2222222222222222222222222222222222222222",\n'
            '            "time_time": "2020-01-01 23:32:08",\n'
            '            "uuid_uuid": "046e4f15-1ddd-5724-b032-878248a71d4b",\n'
            '            "size_size": "1.0TiB",\n'
            '            "strlist_strlist": [\n'
            '                "list",\n'
            '                "string"\n'
            '            ]\n'
            '        }\n'
            '    ]\n'
            '}\n'
        )

        pr = Report(
            _test_obj_types,
            [pf_name, pf_num, pf_sha, pf_time, pf_uuid, pf_size, pf_str_list],
            "name,number,sha,time,uuid,size,strlist",
            opts,
            None,
            "Test",
        )

        for obj in _report_objs:
            pr.report_object(obj)
        pr.report_output()
        print(output.getvalue())
        self.assertEqual(output.getvalue(), xoutput)

    def test_FieldType_with_help(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )

        output = StringIO()
        opts = ReportOpts(report_file=output)

        xoutput = (
            "Str Fields\n"
            "----------\n"
            "  name        - Nothing [str]\n"
            " \n"
            "Num Fields\n"
            "----------\n"
            "  number      - Nothing [num]\n"
        )

        pr = Report(_test_obj_types, [pf_name, pf_num], "help", opts, None, None)

        for obj in _report_objs:
            pr.report_object(obj)
        pr.report_output()

        self.assertEqual(output.getvalue(), xoutput)

    def test_FieldType_bad_field(self):
        pf_name = FieldType(
            PR_STR,
            "name",
            "Name",
            "Nothing",
            8,
            REP_STR,
            lambda f, d: f.report_str(d),
        )
        pf_num = FieldType(
            PR_NUM,
            "number",
            "Number",
            "Nothing",
            8,
            REP_NUM,
            lambda f, d: f.report_num(d),
        )

        output = StringIO()
        opts = ReportOpts(report_file=output)

        with self.assertRaises(ValueError) as cm:
            Report(_test_obj_types, [pf_name, pf_num], "nosuchfield", opts, None, None)

    def test_ReportOpts_str(self):
        opts = ReportOpts()
        xstr = (
            "columns=80\n"
            "headings=True\n"
            "buffered=True\n"
            "separator= \n"
            "field_name_prefix=\n"
            "unquoted=True\n"
            "aligned=True\n"
            "json=False\n"
            "columns_as_rows=False\n"
        )
        self.assertTrue(str(opts).startswith(xstr))

    def test_ReportOpts_eq_equal(self):
        opts1 = ReportOpts()
        opts2 = ReportOpts()
        self.assertEqual(opts1, opts2)

    def test_ReportOpts_eq_notequal(self):
        opts1 = ReportOpts()
        opts2 = ReportOpts(headings=False)
        self.assertNotEqual(opts1, opts2)

    def test_ReportOpts_eq_notsame(self):
        opts = ReportOpts()
        obj = object()
        self.assertNotEqual(opts, obj)

    def test_ReportObjType_no_objtype(self):
        with self.assertRaises(ValueError) as cm:
            ReportObjType(None, "description", "prefix_", lambda x: x)

    def test_ReportObjType_bad_objtype(self):
        with self.assertRaises(ValueError) as cm:
            ReportObjType(-2, "description", "prefix_", lambda x: x)

    def test_ReportObjType_no_desc(self):
        with self.assertRaises(ValueError) as cm:
            ReportObjType(2, None, "prefix_", lambda x: x)

    def test_ReportObjType_no_data_fn(self):
        with self.assertRaises(ValueError) as cm:
            ReportObjType(2, "description", "prefix_", None)


# vim: set et ts=4 sw=4 :



import tempfile
import subprocess
import logging
from pathlib import Path
import os

log = logging.getLogger()

_LOSETUP_CMD = "losetup"

_UDEVADM = "udevadm"
_SETTLE = "settle"

_VGCREATE_CMD = "vgcreate"
_LVCREATE_CMD = "lvcreate"
_VGREMOVE_CMD = "vgremove"
_VGCHANGE_CMD = "vgchange"
_LVS_CMD = "lvs"
_VGS_CMD = "vgs"
_LVM_CMD = "lvm"
_LVM_VERSION = "version"
_LVM_VERSION_STR = "LVM version"

_MOUNT_CMD = "mount"
_UMOUNT_CMD = "umount"

_MKFS_EXT4_CMD = "mkfs.ext4"

# LVM2
_VG_NAME = "test_vg0"

_THIN_POOL_NAME = "pool0"

# 1GiB
_LV_SIZE = 1024**3

# 512MiB
_SNAPSHOT_SIZE = 512 * 2**20

# GiB
_THIN_POOL_SIZE = 6

# Stratis
_STRATIS_CMD = "stratis"
_POOL_CMD = "pool"
_FILESYSTEM_CMD = "filesystem"
_CREATE_CMD = "create"
_DESTROY_CMD = "destroy"
_LIST_CMD = "list"
_START_CMD = "start"
_STOP_CMD = "stop"
_NAME_ARG = "--name"
_SIZE_ARG = "--size"

_POOL_NAME = "pool1"

_FS_SIZE = "1GiB"

# Path in which to create temporary files
_VAR_TMP = "/var/tmp"

# 20GiB
_LOOP_DEVICE_SIZE = 20 * 2**30


def _run(cmd):
    """
    Run command ``cmd`` using ``subprocess.run`` and return the output (if
    any) as a string. A CalledProcessError is raised if the command exits
    with failure.
    """
    log.debug("Calling: '%s'", " ".join(cmd))
    try:
        res = subprocess.run(
            cmd, check=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
    except subprocess.CalledProcessError as err:
        log.error("Command failed: %s", " ".join(cmd))
        log.error("Output: %s", err.output.decode("utf8"))
        raise err
    return res.stdout.decode("utf8").strip()


def _version_string_to_tuple(version):
    return tuple(map(int, version.split(".")))


def _get_lvm_version():
    lvm_version_info = _run([_LVM_CMD, _LVM_VERSION])
    for line in lvm_version_info.splitlines():
        if _LVM_VERSION_STR in line:
            log.debug("Found %s", line)
            (_, _, version, _) = line.split()
            if "(" in version:
                version, _ = version.split("(", maxsplit=1)
            return _version_string_to_tuple(version)
    return _version_string_to_tuple("0.0.0")


def _lvm_supports_devices_file():
    version = _get_lvm_version()
    return version >= (2, 3, 12)


if _lvm_supports_devices_file():
    os.environ["LVM_SYSTEM_DIR"] = os.path.join(os.getcwd(), "tests/lvm")
else:
    os.environ["LVM_SYSTEM_DIR"] = os.path.join(os.getcwd(), "tests/lvm-nodevicesfile")


class LoopBackDevices(object):
    """
    Class for creating and managing Linux loop back devices for testing.
    """

    def __init__(self):
        self.dir = tempfile.mkdtemp("_snapm_loop_back", dir=_VAR_TMP)
        self.count = 0
        self.devices = []

    def create_devices(self, count):
        for index in range(count):
            backing_file = os.path.join(self.dir, f"image{index}")

            with open(backing_file, "ab") as file:
                file.truncate(_LOOP_DEVICE_SIZE)

            device = _run([_LOSETUP_CMD, "--find", "--show", backing_file])
            self.devices.append((device, backing_file))
        self.count = count

    def destroy_devices(self):
        for device, backing_file in self.devices:
            _run([_LOSETUP_CMD, "--detach", device])
            os.remove(backing_file)
        self.devices = []
        self.count = 0

    def destroy_all(self):
        self.destroy_devices()
        os.rmdir(self.dir)

    def device_nodes(self):
        return [dev[0] for dev in self.devices]


class LvmLoopBacked(object):
    """
    Class to create LVM2 volume groups using loop devices.
    """

    def __init__(self, volumes, thin_volumes=None):
        """
        Initialize a new test volume group named ``_VG_NAME``

        :param volumes: A list of linear volumes to create. Each volume
        will be of ``_LV_SIZE``.
        :param thin_volumes: A list of thin volumes to create in the
                             volume group. Each volume will have a
                             virtual size of ``_LV_SIZE``.
        """
        self.mount_root = tempfile.mkdtemp("_snapm_mounts", dir=_VAR_TMP)
        log.debug("Created LVM2 mount_root at %s", self.mount_root)

        # Create loopback device to back VG
        self._lb = LoopBackDevices()
        self._lb.create_devices(1)

        self._create_volume_group()

        if thin_volumes:
            self._create_thin_pool()

        self._create_and_mount_volumes(volumes, thin=False)
        self.volumes = volumes

        self._create_and_mount_volumes(thin_volumes, thin=True)
        self.thin_volumes = thin_volumes

    def _create_volume_group(self):
        _vgcreate_cmd = [_VGCREATE_CMD, _VG_NAME]
        _vgcreate_cmd.extend(self._lb.device_nodes())
        _run(_vgcreate_cmd)

    def _create_and_mount_volumes(self, volumes, thin=False):
        for lv in volumes:
            os.makedirs(os.path.join(self.mount_root, lv))
            self._lvcreate(lv, thin=thin)
            self.format(lv)
            self.mount(lv)

    def _lvcreate(self, name, thin=False):
        if not thin:
            _run(
                [_LVCREATE_CMD, _SIZE_ARG, f"{_LV_SIZE}b", "--name", name, _VG_NAME]
            )
        else:
            _run(
                [
                    _LVCREATE_CMD,
                    "--virtualsize",
                    f"{_LV_SIZE}b",
                    "--thin",
                    "--name",
                    name,
                    f"{_VG_NAME}/{_THIN_POOL_NAME}",
                ]
            )

    def _create_thin_pool(self):
        """
        Create a thin pool of ``_THIN_POOL_SIZE``. The pool will be
        named ``_THIN_POOL_NAME``.
        """
        _run(
            [
                _LVCREATE_CMD,
                _SIZE_ARG,
                f"{_THIN_POOL_SIZE}g",
                "--thin",
                "--name",
                _THIN_POOL_NAME,
                _VG_NAME,
            ]
        )

    def _create_cow_snapshot(self, origin, name):
        """
        Create an LVM2 CoW snapshot in the test VG.
        """
        _run(
            [
                _LVCREATE_CMD,
                "--snapshot",
                _SIZE_ARG,
                f"{_SNAPSHOT_SIZE}b",
                "--name",
                f"{name}",
                f"{_VG_NAME}/{origin}",
            ]
        )

    def _create_thin_snapshot(self, origin, name):
        """
        Create an LVM2 thin snapshot in the test VG.
        """
        _run(
            [
                _LVCREATE_CMD,
                "--snapshot",
                "--name",
                f"{name}",
                f"{_VG_NAME}/{origin}",
            ]
        )

    def create_snapshot(self, origin, name):
        """
        Create a snapshot of volume `origin` in the test VG.
        """
        if origin in self.volumes:
            return self._create_cow_snapshot(origin, name)
        elif origin in self.thin_volumes:
            return self._create_thin_snapshot(origin, name)
        raise ValueError(f"Unknown origin: {origin}")

    def all_volumes(self):
        return self.volumes + self.thin_volumes

    def format(self, name):
        _run([_MKFS_EXT4_CMD, f"/dev/{_VG_NAME}/{name}"])

    def make_mount_point(self, name):
        os.makedirs(os.path.join(self.mount_root, name))

    def mount(self, name):
        _run(
            [_MOUNT_CMD, f"/dev/{_VG_NAME}/{name}", f"{self.mount_root}/{name}"]
        )

    def mount_all(self):
        for lv in self.all_volumes():
            self.mount(lv)

    def umount(self, name):
        _run([_UMOUNT_CMD, f"{self.mount_root}/{name}"])

    def umount_all(self):
        for lv in self.all_volumes():
            self.umount(lv)

    def activate(self):
        _run([_VGCHANGE_CMD, "-ay", _VG_NAME])

    def deactivate(self):
        _run([_VGCHANGE_CMD, "-an", _VG_NAME])

    def mount_points(self):
        return [f"{self.mount_root}/{name}" for name in self.all_volumes()]

    def block_devs(self):
        return [f"/dev/{_VG_NAME}/{name}" for name in self.all_volumes()]

    def touch_path(self, relpath):
        path = Path(f"{self.mount_root}/{relpath}")
        path.touch()

    def test_path(self, relpath):
        path = Path(f"{self.mount_root}/{relpath}")
        return path.exists()

    def dump_lvs(self):
        subprocess.check_call([_LVS_CMD, "-a"])

    def dump_vgs(self):
        subprocess.check_call([_VGS_CMD])

    def destroy(self):
        self.umount_all()
        for lv in self.all_volumes():
            os.rmdir(f"{self.mount_root}/{lv}")
        for subdir in os.listdir(self.mount_root):
            os.rmdir(os.path.join(self.mount_root, subdir))
        os.rmdir(self.mount_root)

        _run([_VGREMOVE_CMD, "--force", "--yes", _VG_NAME])

        self._lb.destroy_all()


class StratisLoopBacked(object):
    """
    Class to create Stratis pools using loop devices.
    """

    def __init__(self, volumes):
        """
        Initialize a new test pool named ``_POOL_NAME``

        :param volumes: A list of linear volumes to create. Each volume
        will be of ``_FS_SIZE``.
        :param thin_volumes: A list of thin volumes to create in the
                             volume group. Each volume will have a
                             virtual size of ``_FS_SIZE``.
        """
        self.mount_root = tempfile.mkdtemp("_snapm_mounts", dir=_VAR_TMP)
        log.debug("Created Stratis mount_root at %s", self.mount_root)

        # Create loopback device to back pool
        self._lb = LoopBackDevices()
        self._lb.create_devices(1)

        self._create_pool()

        self._create_and_mount_volumes(volumes)
        self.volumes = volumes

    def _create_pool(self):
        _stratis_pool_create_cmd = [_STRATIS_CMD, _POOL_CMD, _CREATE_CMD, _POOL_NAME]
        _stratis_pool_create_cmd.extend(self._lb.device_nodes())
        _run(_stratis_pool_create_cmd)

    def _create_and_mount_volumes(self, volumes):
        for fs in volumes:
            os.makedirs(os.path.join(self.mount_root, fs))
            self._fs_create(fs)
            _run(
                [
                    _UDEVADM,
                    _SETTLE,
                ]
            )
            self.mount(fs)

    def _fs_create(self, name, thin=False):
        _run(
            [
                _STRATIS_CMD,
                _FILESYSTEM_CMD,
                _CREATE_CMD,
                _SIZE_ARG,
                _FS_SIZE,
                _POOL_NAME,
                name,
            ]
        )

    def create_snapshot(self, origin, name):
        """
        Create a snapshot of volume `origin` in the test pool.
        """
        if origin in self.volumes:
            return _run(
                [
                    _STRATIS_CMD,
                    _FILESYSTEM_CMD,
                    "snapshot",
                    _POOL_NAME,
                    origin,
                    name,
                ]
            )
        raise ValueError(f"Unknown origin: {origin}")

    def all_volumes(self):
        return self.volumes

    def mount(self, name):
        _run(
            [_MOUNT_CMD, f"/dev/stratis/{_POOL_NAME}/{name}", f"{self.mount_root}/{name}"]
        )

    def mount_all(self):
        for fs in self.all_volumes():
            self.mount(fs)

    def umount(self, name):
        _run([_UMOUNT_CMD, f"{self.mount_root}/{name}"])

    def umount_all(self):
        for fs in self.all_volumes():
            self.umount(fs)

    def mount_points(self):
        return [f"{self.mount_root}/{name}" for name in self.all_volumes()]

    def block_devs(self):
        return [f"/dev/stratis/{_POOL_NAME}/{name}" for name in self.all_volumes()]

    def touch_path(self, relpath):
        path = Path(f"{self.mount_root}/{relpath}")
        path.touch()

    def test_path(self, relpath):
        path = Path(f"{self.mount_root}/{relpath}")
        return path.exists()

    def dump_fs(self):
        subprocess.check_call([_STRATIS_CMD, _FILESYSTEM_CMD])

    def dump_pools(self):
        subprocess.check_call([_STRATIS_CMD, _POOL_CMD])

    def start_pool(self):
        _run([_STRATIS_CMD, _POOL_CMD, _START_CMD, _NAME_ARG, _POOL_NAME])
        _run(["udevadm", "settle"])

    def stop_pool(self):
        _run([_STRATIS_CMD, _POOL_CMD, _STOP_CMD, _NAME_ARG, _POOL_NAME])

    def _destroy_one_filesystem(self, pool, filesystem):
        _run(
            [
                _STRATIS_CMD,
                _FILESYSTEM_CMD,
                _DESTROY_CMD,
                pool,
                filesystem,
            ]
        )

    def _destroy_all_filesystems(self):
        stratis_filesystem_cmd_args = [
            _STRATIS_CMD,
            _FILESYSTEM_CMD,
            _LIST_CMD,
            _POOL_NAME,
        ]

        stratis_filesystem_out = _run(
            stratis_filesystem_cmd_args,
        )

        for line in stratis_filesystem_out.splitlines():
            if "Pool" in line and "Filesystem" in line:
                continue
            fields = line.split()
            self._destroy_one_filesystem(_POOL_NAME, fields[1])

    def destroy(self):
        self.umount_all()
        for fs in self.all_volumes():
            os.rmdir(f"{self.mount_root}/{fs}")
        os.rmdir(self.mount_root)

        for fs in self.all_volumes():
            self._destroy_one_filesystem(_POOL_NAME, fs)

        # Cleanup left-over file systems
        self._destroy_all_filesystems()

        _run([_STRATIS_CMD, _POOL_CMD, _DESTROY_CMD, _POOL_NAME])

        self._lb.destroy_all()



# Copyright Red Hat
#
# snapm/manager/plugins/__init__.py - Snapshot Manager plugins
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
"""
Snapshot manager plugin interface.
"""
from ._plugin import *  # noqa: F401, F403
from ._plugin import __all__  # noqa: F401



# Copyright Red Hat
#
# snapm/command.py - Snapshot manager command interface
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
"""The ``snapm.command`` module provides both the snapm command line
interface infrastructure, and a simple procedural interface to the
``snapm`` library modules.

The procedural interface is used by the ``snapm`` command line tool,
and may be used by application programs, or interactively in the
Python shell by users who do not require all the features present
in the snapm object API.
"""
from argparse import ArgumentParser
from os.path import basename
from json import dumps
from uuid import UUID
import logging

from snapm import (
    SnapmInvalidIdentifierError,
    SNAPSET_NAME,
    SNAPSET_SOURCES,
    SNAPSET_MOUNT_POINTS,
    SNAPSET_DEVICES,
    SNAPSET_NR_SNAPSHOTS,
    SNAPSET_TIME,
    SNAPSET_TIMESTAMP,
    SNAPSET_UUID,
    SNAPSET_STATUS,
    SNAPSET_AUTOACTIVATE,
    SNAPSET_BOOTABLE,
    SNAPSET_SNAPSHOT_ENTRY,
    SNAPSET_REVERT_ENTRY,
    SNAPSHOT_NAME,
    SNAPSHOT_ORIGIN,
    SNAPSHOT_SOURCE,
    SNAPSHOT_MOUNT_POINT,
    SNAPSHOT_PROVIDER,
    SNAPSHOT_UUID,
    SNAPSHOT_STATUS,
    SNAPSHOT_SIZE,
    SNAPSHOT_FREE,
    SNAPSHOT_SIZE_BYTES,
    SNAPSHOT_FREE_BYTES,
    SNAPSHOT_AUTOACTIVATE,
    SNAPSHOT_DEV_PATH,
    SNAPM_DEBUG_MANAGER,
    SNAPM_DEBUG_COMMAND,
    SNAPM_DEBUG_REPORT,
    SNAPM_DEBUG_ALL,
    set_debug_mask,
    bool_to_yes_no,
    Selection,
    __version__,
)
from snapm.manager import Manager
from snapm.report import (
    REP_NUM,
    REP_SHA,
    REP_STR,
    REP_TIME,
    REP_UUID,
    REP_SIZE,
    REP_STR_LIST,
    ReportOpts,
    ReportObjType,
    FieldType,
    Report,
)

_log = logging.getLogger(__name__)
_log.set_debug_mask(SNAPM_DEBUG_COMMAND)

_log_debug = _log.debug
_log_debug_command = _log.debug_masked
_log_info = _log.info
_log_warn = _log.warning
_log_error = _log.error

_DEFAULT_LOG_LEVEL = logging.WARNING
_CONSOLE_HANDLER = None

#
# Reporting object types
#


class ReportObj:
    """
    Common report object for snapm reports
    """

    def __init__(self, snapset=None, snapshot=None, plugin=None):
        self.snapset = snapset
        self.snapshot = snapshot
        self.plugin = plugin


#: Snapshot set report object type
PR_SNAPSET = 1
#: Snapshot report object type
PR_SNAPSHOT = 2
#: Plugin report object type
PR_PLUGIN = 4

#: Report object types table for ``snapm.command`` reports
_report_obj_types = [
    ReportObjType(PR_SNAPSET, "Snapshot set", "snapset_", lambda o: o.snapset),
    ReportObjType(PR_SNAPSHOT, "Snapshot", "snapshot_", lambda o: o.snapshot),
    ReportObjType(PR_PLUGIN, "Plugin", "plugin_", lambda o: o.plugin),
]


#
# Report field definitions
#

_snapshot_set_fields = [
    FieldType(
        PR_SNAPSET,
        "name",
        SNAPSET_NAME,
        "Snapshot set name",
        12,
        REP_STR,
        lambda f, d: f.report_str(d.name),
    ),
    FieldType(
        PR_SNAPSET,
        "uuid",
        SNAPSET_UUID,
        "Snapshot set UUID",
        37,
        REP_UUID,
        lambda f, d: f.report_uuid(d.uuid),
    ),
    FieldType(
        PR_SNAPSET,
        "timestamp",
        SNAPSET_TIMESTAMP,
        "Snapshot set creation time as a UNIX epoch value",
        10,
        REP_NUM,
        lambda f, d: f.report_num(d.timestamp),
    ),
    FieldType(
        PR_SNAPSET,
        "time",
        SNAPSET_TIME,
        "Snapshot set creation time",
        20,
        REP_TIME,
        lambda f, d: f.report_time(d.time),
    ),
    FieldType(
        PR_SNAPSET,
        "nr_snapshots",
        SNAPSET_NR_SNAPSHOTS,
        "Number of snapshots",
        11,
        REP_NUM,
        lambda f, d: f.report_num(d.nr_snapshots),
    ),
    FieldType(
        PR_SNAPSET,
        "sources",
        SNAPSET_SOURCES,
        "Snapshot set sources",
        8,
        REP_STR_LIST,
        lambda f, d: f.report_str_list(d.sources),
    ),
    FieldType(
        PR_SNAPSET,
        "mountpoints",
        SNAPSET_MOUNT_POINTS,
        "Snapshot set mount points",
        24,
        REP_STR_LIST,
        lambda f, d: f.report_str_list(d.mount_points),
    ),
    FieldType(
        PR_SNAPSET,
        "devices",
        SNAPSET_DEVICES,
        "Snapshot set devices",
        9,
        REP_STR_LIST,
        lambda f, d: f.report_str_list(d.devices),
    ),
    FieldType(
        PR_SNAPSET,
        "status",
        SNAPSET_STATUS,
        "Snapshot set status",
        7,
        REP_STR,
        lambda f, d: f.report_str(str(d.status)),
    ),
    FieldType(
        PR_SNAPSET,
        "autoactivate",
        SNAPSET_AUTOACTIVATE,
        "Autoactivation status",
        12,
        REP_STR,
        lambda f, d: f.report_str(bool_to_yes_no(d.autoactivate)),
    ),
    FieldType(
        PR_SNAPSET,
        "bootable",
        SNAPSET_BOOTABLE,
        "Configured for snapshot boot",
        8,
        REP_STR,
        lambda f, d: f.report_str(bool_to_yes_no(d.boot_entry is not None)),
    ),
    FieldType(
        PR_SNAPSET,
        "bootentry",
        SNAPSET_SNAPSHOT_ENTRY,
        "Snapshot set boot entry",
        13,
        REP_SHA,
        lambda f, d: f.report_sha("" if not d.boot_entry else d.boot_entry.boot_id),
    ),
    FieldType(
        PR_SNAPSET,
        "revertentry",
        SNAPSET_REVERT_ENTRY,
        "Snapshot set revert boot entry",
        13,
        REP_SHA,
        lambda f, d: f.report_sha("" if not d.revert_entry else d.revert_entry.boot_id),
    ),
]

_DEFAULT_SNAPSET_FIELDS = "name,time,nr_snapshots,status,sources"

_snapshot_fields = [
    FieldType(
        PR_SNAPSHOT,
        "name",
        SNAPSHOT_NAME,
        "Snapshot name",
        24,
        REP_STR,
        lambda f, d: f.report_str(d.name),
    ),
    FieldType(
        PR_SNAPSHOT,
        "uuid",
        SNAPSHOT_UUID,
        "Snapshot UUID",
        37,
        REP_UUID,
        lambda f, d: f.report_uuid(d.uuid),
    ),
    FieldType(
        PR_SNAPSHOT,
        "origin",
        "Origin",
        SNAPSHOT_ORIGIN,
        16,
        REP_STR,
        lambda f, d: f.report_str(d.origin),
    ),
    FieldType(
        PR_SNAPSHOT,
        "source",
        SNAPSHOT_SOURCE,
        "Snapshot source",
        7,
        REP_STR,
        lambda f, d: f.report_str(d.source),
    ),
    FieldType(
        PR_SNAPSHOT,
        "mountpoint",
        SNAPSHOT_MOUNT_POINT,
        "Snapshot mount point",
        16,
        REP_STR,
        lambda f, d: f.report_str(d.mount_point),
    ),
    FieldType(
        PR_SNAPSHOT,
        "devpath",
        SNAPSHOT_DEV_PATH,
        "Snapshot device path",
        8,
        REP_STR,
        lambda f, d: f.report_str(d.devpath),
    ),
    FieldType(
        PR_SNAPSHOT,
        "provider",
        SNAPSHOT_PROVIDER,
        "Snapshot provider plugin",
        8,
        REP_STR,
        lambda f, d: f.report_str(d.provider.name),
    ),
    FieldType(
        PR_SNAPSHOT,
        "status",
        SNAPSHOT_STATUS,
        "Snapshot status",
        7,
        REP_STR,
        lambda f, d: f.report_str(str(d.status)),
    ),
    FieldType(
        PR_SNAPSHOT,
        "size",
        SNAPSHOT_SIZE,
        "Snapshot size",
        6,
        REP_SIZE,
        lambda f, d: f.report_size(d.size),
    ),
    FieldType(
        PR_SNAPSHOT,
        "free",
        SNAPSHOT_FREE,
        "Free space available",
        6,
        REP_SIZE,
        lambda f, d: f.report_size(d.free),
    ),
    FieldType(
        PR_SNAPSHOT,
        "size_bytes",
        SNAPSHOT_SIZE_BYTES,
        "Snapshot size in bytes",
        6,
        REP_NUM,
        lambda f, d: f.report_num(d.size),
    ),
    FieldType(
        PR_SNAPSHOT,
        "free_bytes",
        SNAPSHOT_FREE_BYTES,
        "Free space available in bytes",
        6,
        REP_NUM,
        lambda f, d: f.report_num(d.free),
    ),
    FieldType(
        PR_SNAPSHOT,
        "autoactivate",
        SNAPSHOT_AUTOACTIVATE,
        "Autoactivation status",
        12,
        REP_STR,
        lambda f, d: f.report_str(bool_to_yes_no(d.autoactivate)),
    ),
]

_DEFAULT_SNAPSHOT_FIELDS = (
    "snapset_name,name,origin,source,status,size,free,autoactivate,provider"
)

_plugin_fields = [
    FieldType(
        PR_PLUGIN,
        "name",
        "PluginName",
        "Name of the plugin",
        10,
        REP_STR,
        lambda f, d: f.report_str(d.name),
    ),
    FieldType(
        PR_PLUGIN,
        "version",
        "PluginVersion",
        "Version of the plugin",
        13,
        REP_STR,
        lambda f, d: f.report_str(d.version),
    ),
    FieldType(
        PR_PLUGIN,
        "type",
        "PluginType",
        "The snapshot type created by this plugin",
        10,
        REP_STR,
        lambda f, d: f.report_str(d.snapshot_class.__name__),
    ),
]

_DEFAULT_PLUGIN_FIELDS = "name,version,type"


def _str_indent(string, indent):
    """
    Indent all lines of a multi-line string.

    Indent each line of the multi line string ``string`` to the
    specified indentation level.

    :param string: The string to be indented
    :param indent: The number of characters to indent by
    :returns: str
    """
    outstr = ""
    for line in string.splitlines():
        outstr += indent * " " + line + "\n"
    return outstr.rstrip("\n")


def _do_print_type(
    report_fields, selected, output_fields=None, opts=None, sort_keys=None, title=None
):
    """
    Print an object type report (snapshot set, snapshot)

    Helper for list function that generate reports.

    Format a set of snapshot set or snapshot objects matching
    the given criteria and format them as a report, returning
    the output as a string.

    Selection criteria may be expressed via a Selection object
    passed to the call using the ``selection`` parameter.

    :param selection: A Selection object giving selection
                      criteria for the operation
    :param output_fields: a comma-separated list of output fields
    :param opts: output formatting and control options
    :param sort_keys: a comma-separated list of sort keys
    :rtype: str
    """
    opts = opts if opts is not None else ReportOpts()

    report = Report(
        _report_obj_types, report_fields, output_fields, opts, sort_keys, title
    )

    for obj in selected:
        report.report_object(obj)

    return report.report_output()


def create_snapset(manager, name, sources, size_policy=None, boot=False, revert=False):
    """
    Create a new snapshot set from a list of mount point and block device
    source paths.

    :param manager: The manager context to use
    :param name: The name of the new snapshot set
    :param sources: A list of mount point or block devices to snapshot
    :param size_policy: The default size policy for this snapshot set.
    :param boot: Create a boot entry for this snapshot set.
    :param revert: Create a revert boot entry for this snapshot set.
    """
    return manager.create_snapshot_set(
        name, sources, default_size_policy=size_policy, boot=boot, revert=revert
    )


def delete_snapset(manager, selection):
    """
    Delete snapshot set matching selection criteria.

    :param manager: The manager context to use
    :param selection: Selection criteria for the snapshot set to remove.
    """
    if not selection.is_single():
        raise SnapmInvalidIdentifierError("Delete requires unique selection criteria")
    return manager.delete_snapshot_sets(selection)


def rename_snapset(manager, old_name, new_name):
    """
    Rename a snapshot set from ``old_name`` to ``new_name``.
    """
    return manager.rename_snapshot_set(old_name, new_name)


def resize_snapset(manager, sources, name=None, uuid=None, default_size_policy=None):
    """
    Resize snapshot set by name or UUID.

    :param manager: The manager context to use
    :param sources: A list of mount point or block devices to snapshot
    :param name: The name of the snapshot set to resize.
    :param uuid: The uuid of the snapshot set to resize.
    :param default_size_policy: The default size policy for this snapshot set.
    """
    return manager.resize_snapshot_set(
        sources, name=name, uuid=uuid, default_size_policy=default_size_policy
    )


def revert_snapset(manager, name=None, uuid=None):
    """
    Revert snapshot set matching selection criteria.

    :param manager: The manager context to use
    :param selection: Selection criteria for the snapshot set to revert.
    """
    return manager.revert_snapshot_set(name=name, uuid=uuid)


def split_snapset(manager, name, new_name, sources):
    """
    Split an existing snapshot set.

    Split the snapshot set named ``name`` into two snapshot sets, including
    all sources listed in ``sources`` in the newly created snapshot set with
    name ``new_name``.

    :param name: The name of the snapshot set to split.
    :param new_name: The name for the newly created snapshot set.
    :param sources: The sources to split from ``name`` to ``new_name``.
    :returns: A ``SnapshotSet`` object representing the snapshot set named
              ``new_name``.
    """
    return manager.split_snapshot_set(name, new_name, sources)


def prune_snapset(manager, name, sources):
    """
    Prune snapshots from an existing snapshot set.

    Remove snapshots from an existing snapshot set named ``name``. The
    snapshot sources listed in ``sources`` are pruned (deleted) from the
    named snapshot set.

    :param name: The name of the snapshot set to prune.
    :param sources: The sources to prune from ``name``.
    :returns: A ``SnapshotSet`` object representing the pruned snapshot set
    """
    return manager.split_snapshot_set(name, None, sources)


def show_snapshots(manager, selection=None, json=False):
    """
    Show snapshots matching selection criteria.
    """
    snapshots = manager.find_snapshots(selection=selection)
    first = True

    if json:
        snap_list = []

    for snapshot in snapshots:
        if json:
            snap_list.append(snapshot.as_dict())
            continue
        wspace = "" if first else "\n"
        print(f"{wspace}{snapshot}")
        first = False

    if json:
        print(dumps(snap_list, indent=4))


def show_snapsets(manager, selection=None, members=False, json=False):
    """
    Show snapshot sets matching selection criteria.
    """
    snapsets = manager.find_snapshot_sets(selection=selection)
    first = True

    if json:
        set_list = []

    for snapset in snapsets:
        if json:
            set_list.append(snapset.as_dict(members=members))
            continue
        wspace = "" if first else "\n"
        print(f"{wspace}{snapset}")
        first = False
        if members:
            print("Snapshots:")
            first_snapshot = True
            for snapshot in snapset.snapshots:
                wspace = "" if first_snapshot else "\n"
                print(wspace + _str_indent(str(snapshot), 4))
                first_snapshot = False

    if json:
        print(dumps(set_list, indent=4))


def _expand_fields(default_fields, output_fields):
    """
    Expand output fields list from command line arguments.
    """

    if not output_fields:
        output_fields = default_fields
    elif output_fields.startswith("+"):
        output_fields = default_fields + "," + output_fields[1:]
    return output_fields


def print_plugins(
    manager, selection=None, output_fields=None, opts=None, sort_keys=None
):
    """
    Print plugins matching selection criteria.

    Format a set of ``snapm.manager.Plugin`` objects matching
    the given criteria, and output them as a report to the file
    given in ``opts.report_file``.

    Selection criteria are currently ignored for plugin reports.

    :param selection: A Selection object giving selection
                      criteria for the operation
    :param output_fields: a comma-separated list of output fields
    :param opts: output formatting and control options
    :param sort_keys: a comma-separated list of sort keys
    """
    del selection
    output_fields = _expand_fields(_DEFAULT_PLUGIN_FIELDS, output_fields)

    plugins = manager.plugins
    selected = [ReportObj(None, None, plugin) for plugin in plugins]

    return _do_print_type(
        _plugin_fields,
        selected,
        output_fields=output_fields,
        opts=opts,
        sort_keys=sort_keys,
        title="Plugins",
    )


def print_snapshots(
    manager, selection=None, output_fields=None, opts=None, sort_keys=None
):
    """
    Print snapshots matching selection criteria.

    Format a set of ``snapm.manager.Snapshot`` objects matching
    the given criteria, and output them as a report to the file
    given in ``opts.report_file``.

    Selection criteria may be expressed via a Selection object
    passed to the call using the ``selection`` parameter.

    :param selection: A Selection object giving selection
                      criteria for the operation
    :param output_fields: a comma-separated list of output fields
    :param opts: output formatting and control options
    :param sort_keys: a comma-separated list of sort keys
    """
    output_fields = _expand_fields(_DEFAULT_SNAPSHOT_FIELDS, output_fields)

    snapshots = manager.find_snapshots(selection=selection)
    selected = [ReportObj(snap.snapshot_set, snap) for snap in snapshots]

    return _do_print_type(
        _snapshot_fields + _snapshot_set_fields,
        selected,
        output_fields=output_fields,
        opts=opts,
        sort_keys=sort_keys,
        title="Snapshots",
    )


def print_snapsets(
    manager, selection=None, output_fields=None, opts=None, sort_keys=None
):
    """
    Print snapshot sets matching selection criteria.

    Format a set of ``snapm.manager.SnapshotSet`` objects matching
    the given criteria, and output them as a report to the file
    given in ``opts.report_file``.

    Selection criteria may be expressed via a Selection object
    passed to the call using the ``selection`` parameter.

    :param selection: A Selection object giving selection
                      criteria for the operation
    :param output_fields: a comma-separated list of output fields
    :param opts: output formatting and control options
    :param sort_keys: a comma-separated list of sort keys
    """
    output_fields = _expand_fields(_DEFAULT_SNAPSET_FIELDS, output_fields)

    snapshot_sets = manager.find_snapshot_sets(selection=selection)
    selected = [ReportObj(ss, None) for ss in snapshot_sets]

    return _do_print_type(
        _snapshot_set_fields,
        selected,
        output_fields=output_fields,
        opts=opts,
        sort_keys=sort_keys,
        title="Snapsets",
    )


def _generic_list_cmd(cmd_args, select, opts, manager, print_fn):
    """
    Generic list command implementation.

    Implements a simple list command that applies selection criteria
    and calls a print_*() API function to display results.

    Callers should initialise identifier and select appropriately
    for the specific command arguments.

    :param cmd_args: the command arguments
    :param select: selection criteria
    :param opts: reporting options object
    :param print_fn: the API call to display results. The function
                     must accept the selection, output_fields,
                     opts, and sort_keys keyword arguments
    :returns: None
    """
    if cmd_args.options:
        fields = cmd_args.options
    else:
        fields = None

    if cmd_args.debug:
        print_fn(
            manager,
            selection=select,
            output_fields=fields,
            opts=opts,
            sort_keys=cmd_args.sort,
        )
    else:
        try:
            print_fn(
                manager,
                selection=select,
                output_fields=fields,
                opts=opts,
                sort_keys=cmd_args.sort,
            )
        except ValueError as err:
            print(err)
            return 1
    return 0


def _create_cmd(cmd_args):
    """
    Create snapshot set command handler.
    Attempt to create the specified snapshot set.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    snapset = create_snapset(
        manager,
        cmd_args.snapset_name,
        cmd_args.sources,
        size_policy=cmd_args.size_policy,
        boot=cmd_args.bootable,
        revert=cmd_args.revert,
    )
    if snapset is None:
        return 1
    _log_info(
        "Created snapset %s with %d snapshots", snapset.name, snapset.nr_snapshots
    )
    print(snapset)
    return 0


def _delete_cmd(cmd_args):
    """
    Delete snapshot set command handler.

    Attempt to delete the specified snapshot set.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    select = Selection.from_cmd_args(cmd_args)
    count = delete_snapset(manager, select)
    _log_info("Deleted %d snapshot set%s", count, "s" if count > 1 else "")
    return 0


def _rename_cmd(cmd_args):
    """
    Rename snapshot set command handler.

    Attempt to rename the specified snapshot set.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    rename_snapset(manager, cmd_args.old_name, cmd_args.new_name)
    _log_info("Renamed snapshot set '%s' to '%s'", cmd_args.old_name, cmd_args.new_name)
    return 0


def _resize_cmd(cmd_args):
    """
    Resize snapshot set command handler.

    Attempt to resize the snapshots contained in the given snapshot set
    according to the corresponding size policy.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    name = None
    uuid = None

    selection = Selection.from_cmd_args(cmd_args)
    if not selection.is_single():
        raise SnapmInvalidIdentifierError("Resize requires unique selection criteria")
    if selection.name:
        name = selection.name
    elif selection.uuid:
        uuid = selection.uuid
    else:
        raise SnapmInvalidIdentifierError("Resize requires a snapset name or UUID")

    resize_snapset(
        manager,
        cmd_args.sources,
        name=name,
        uuid=uuid,
        default_size_policy=cmd_args.size_policy,
    )
    return 0


def _revert_cmd(cmd_args):
    """
    Delete snapshot set command handler.

    Attempt to revert the specified snapshot set.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    name = None
    uuid = None

    selection = Selection.from_cmd_args(cmd_args)
    if not selection.is_single():
        raise SnapmInvalidIdentifierError("Revert requires unique selection criteria")
    if selection.name:
        name = selection.name
    elif selection.uuid:
        uuid = selection.uuid
    else:
        raise SnapmInvalidIdentifierError("Revert requires a snapset name or UUID")

    snapset = revert_snapset(manager, name=name, uuid=uuid)
    _log_info("Started revert for snapset %s", snapset.name)
    return 0


def _split_cmd(cmd_args):
    """
    Split snapshot set command handler.

    Attempt to split the specified sources from the first snapshot set
    argument into the second.

    :param cmd_args: Command line arguments for the command.
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()

    snapset = split_snapset(manager, cmd_args.name, cmd_args.new_name, cmd_args.sources)
    _log_info(
        "Split snapset '%s' into '%s' (%s)",
        cmd_args.name,
        snapset.name,
        ", ".join(snapset.sources),
    )
    print(snapset)
    return 0


def _prune_cmd(cmd_args):
    """
    Prune snapshot set command handler.

    Attempt to prune the specified sources from the given snapshot set.

    :param cmd_args: Command line arguments for the command.
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()

    snapset = prune_snapset(manager, cmd_args.name, cmd_args.sources)
    _log_info(
        "Pruned sources (%s) from snapset '%s'",
        snapset.name,
        ", ".join(cmd_args.sources),
    )
    print(snapset)
    return 0


def _activate_cmd(cmd_args):
    """
    Activate snapshot set command handler.

    Attempt to activate the snapshot sets that match the given
    selection criteria.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    select = Selection.from_cmd_args(cmd_args)
    count = manager.activate_snapshot_sets(select)
    _log_info("Activated %d snapshot sets", count)
    return 0


def _deactivate_cmd(cmd_args):
    """
    Deactivate snapshot set command handler.

    Attempt to deactivate the snapshot sets that match the given
    selection criteria.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    select = Selection.from_cmd_args(cmd_args)
    count = manager.deactivate_snapshot_sets(select)
    _log_info("Deactivated %d snapshot sets", count)
    return 0


def _autoactivate_cmd(cmd_args):
    """
    Autoactivation status snapshot set command handler.

    Attempt to set the autoactivation status for snapshot set that match
    the given selection criteria.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    select = Selection.from_cmd_args(cmd_args)
    auto = bool(cmd_args.yes)
    count = manager.set_autoactivate(select, auto=auto)
    _log_info("Set autoactivate=%s for %d snapshot sets", bool_to_yes_no(auto), count)
    return 0


def _list_cmd(cmd_args):
    """
    List snapshot sets command handler.

    List the snapshot sets that match the given selection criteria as
    a tabular report, with one snapshot set per row.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    opts = _report_opts_from_args(cmd_args)
    select = Selection.from_cmd_args(cmd_args)
    return _generic_list_cmd(cmd_args, select, opts, manager, print_snapsets)


def _show_cmd(cmd_args):
    """
    Show snapshot set command handler.

    Show the snapshot sets that match the given selection criteria as
    a multi-line report.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    select = Selection.from_cmd_args(cmd_args)
    show_snapsets(
        manager, selection=select, members=cmd_args.members, json=cmd_args.json
    )
    return 0


def _snapshot_activate_cmd(cmd_args):
    """
    Activate snapshot command handler.

    Attempt to activate the snapshots that match the given
    selection criteria.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    select = Selection.from_cmd_args(cmd_args)
    snapshots = manager.find_snapshots(select)
    if not snapshots:
        _log_error("Could not find snapshots matching %s", select)
        return 1
    count = 0
    for snapshot in snapshots:
        snapshot.activate()
        count += 1
    _log_info("Activated %d snapshots", count)
    return 0


def _snapshot_deactivate_cmd(cmd_args):
    """
    Deactivate snapshot command handler.

    Attempt to deactivate the snapshots that match the given
    selection criteria.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    select = Selection.from_cmd_args(cmd_args)
    snapshots = manager.find_snapshots(select)
    if not snapshots:
        _log_error("Could not find snapshots matching %s", select)
        return 1
    count = 0
    for snapshot in snapshots:
        snapshot.deactivate()
        count += 1
    _log_info("Deactivated %d snapshots", count)
    return 0


def _snapshot_autoactivate_cmd(cmd_args):
    """
    Autoactivate snapshot command handler.

    Attempt to set autoactivation status for the snapshots that match
    the given selection criteria.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    select = Selection.from_cmd_args(cmd_args)
    snapshots = manager.find_snapshots(select)
    if not snapshots:
        _log_error("Could not find snapshots matching %s", select)
        return 1
    count = 0
    auto = bool(cmd_args.yes)
    for snapshot in snapshots:
        snapshot.autoactivate = auto
        count += 1
    _log_info("Set autoactivation status for %d snapshots", count)
    return 0


def _snapshot_list_cmd(cmd_args):
    """
    List snapshots command handler.

    List the snapshot that match the given selection criteria as
    a tabular report, with one snapshot per row.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    opts = _report_opts_from_args(cmd_args)
    select = Selection.from_cmd_args(cmd_args)
    return _generic_list_cmd(cmd_args, select, opts, manager, print_snapshots)


def _snapshot_show_cmd(cmd_args):
    """
    Show snapshots command handler.

    Show the snapshots that match the given selection criteria as
    a multi-line report.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    select = Selection.from_cmd_args(cmd_args)
    show_snapshots(manager, selection=select, json=cmd_args.json)
    return 0


def _plugin_list_cmd(cmd_args):
    """
    List available plugins.

    :param cmd_args: Command line arguments for the command
    :returns: integer status code returned from ``main()``
    """
    manager = Manager()
    opts = _report_opts_from_args(cmd_args)
    return _generic_list_cmd(cmd_args, None, opts, manager, print_plugins)


def _report_opts_from_args(cmd_args):
    opts = ReportOpts()

    if not cmd_args:
        return opts

    if cmd_args.json:
        opts.json = True

    if cmd_args.rows:
        opts.columns_as_rows = True

    if cmd_args.separator:
        opts.separator = cmd_args.separator

    if cmd_args.name_prefixes:
        opts.field_name_prefix = "SNAPM_"
        opts.unquoted = False
        opts.aligned = False

    if cmd_args.no_headings:
        opts.headings = False

    return opts


def setup_logging(cmd_args):
    """
    Set up snapm logging.
    """
    # pylint: disable=global-statement
    global _CONSOLE_HANDLER
    level = _DEFAULT_LOG_LEVEL
    if cmd_args.verbose and cmd_args.verbose > 1:
        level = logging.DEBUG
    elif cmd_args.verbose and cmd_args.verbose > 0:
        level = logging.INFO
    snapm_log = logging.getLogger("snapm")
    snapm_log.setLevel(level)
    formatter = logging.Formatter("%(levelname)s - %(message)s")
    if _CONSOLE_HANDLER not in snapm_log.handlers:
        _CONSOLE_HANDLER = logging.StreamHandler()
        snapm_log.addHandler(_CONSOLE_HANDLER)
    _CONSOLE_HANDLER.setLevel(level)
    _CONSOLE_HANDLER.setFormatter(formatter)


def shutdown_logging():
    """
    Shut down snapm logging.
    """
    logging.shutdown()


def set_debug(debug_arg):
    """
    Set debugging mask from command line argument.
    """
    if not debug_arg:
        return

    mask_map = {
        "manager": SNAPM_DEBUG_MANAGER,
        "command": SNAPM_DEBUG_COMMAND,
        "report": SNAPM_DEBUG_REPORT,
        "all": SNAPM_DEBUG_ALL,
    }

    mask = 0
    for name in debug_arg.split(","):
        if name not in mask_map:
            raise ValueError(f"Unknown debug option: {name}")
        mask |= mask_map[name]
    set_debug_mask(mask)


def _add_identifier_args(parser, snapset=False, snapshot=False):
    """
    Add snapshot set or snapshot identifier command line arguments.
    """
    if snapset:
        parser.add_argument(
            "-n",
            "--name",
            metavar="NAME",
            type=str,
            help="A snapset name",
        )
        parser.add_argument(
            "-u",
            "--uuid",
            metavar="UUID",
            type=UUID,
            help="A snapset UUID",
        )
    if snapshot:
        parser.add_argument(
            "-N",
            "--snapshot-name",
            metavar="SNAPSHOT_NAME",
            type=str,
            help="A snapshot name",
        )
        parser.add_argument(
            "-U",
            "--snapshot-uuid",
            metavar="SNAPSHOT_UUID",
            type=UUID,
            help="A snapshot UUID",
        )
    parser.add_argument(
        "identifier",
        metavar="ID",
        type=str,
        action="store",
        help="An optional snapset name or UUID to operate on",
        nargs="?",
        default=None,
    )


def _add_report_args(parser):
    """
    Add common reporting command line arguments.
    """
    parser.add_argument(
        "--name-prefixes",
        "--nameprefixes",
        help="Add a prefix to report field names",
        action="store_true",
    )
    parser.add_argument(
        "--no-headings",
        "--noheadings",
        action="store_true",
        help="Suppress output of report headings",
    )
    parser.add_argument(
        "-o",
        "--options",
        metavar="FIELDS",
        type=str,
        help="Specify which fields to display",
    )
    parser.add_argument(
        "-O",
        "--sort",
        metavar="SORTFIELDS",
        type=str,
        help="Specify which fields to sort by",
    )
    format_group = parser.add_mutually_exclusive_group()
    format_group.add_argument(
        "--json", action="store_true", help="Output report as JSON"
    )
    format_group.add_argument(
        "--rows", action="store_true", help="Output report columnes as rows"
    )
    parser.add_argument(
        "--separator", metavar="SEP", type=str, help="Report field separator"
    )


def _add_autoactivate_args(parser):
    parser.add_argument(
        "--yes",
        action="store_true",
        help="Enable snapshot autoactivation",
    )
    parser.add_argument(
        "--no",
        action="store_true",
        help="Disable snapshot autoactivation",
    )


def _add_json_arg(parser):
    parser.add_argument(
        "-j",
        "--json",
        action="store_true",
        help="Display output in JSON notation",
    )


CREATE_CMD = "create"
DELETE_CMD = "delete"
RENAME_CMD = "rename"
RESIZE_CMD = "resize"
REVERT_CMD = "revert"
SPLIT_CMD = "split"
PRUNE_CMD = "prune"
ACTIVATE_CMD = "activate"
DEACTIVATE_CMD = "deactivate"
AUTOACTIVATE_CMD = "autoactivate"
SHOW_CMD = "show"
LIST_CMD = "list"

SNAPSET_TYPE = "snapset"
SNAPSHOT_TYPE = "snapshot"
PLUGIN_TYPE = "plugin"


# pylint: disable=too-many-statements
def _add_snapset_subparser(type_subparser):
    """
    Add subparser for 'snapset' commands.

    :param type_subparser: Command type subparser
    """
    # Subparser for snapset commands
    snapset_parser = type_subparser.add_parser(
        SNAPSET_TYPE, help="Snapshot set commands"
    )
    snapset_subparser = snapset_parser.add_subparsers(dest="command")

    # snapset create subcommand
    snapset_create_parser = snapset_subparser.add_parser(
        CREATE_CMD, help="Create snapshot sets"
    )
    snapset_create_parser.set_defaults(func=_create_cmd)
    snapset_create_parser.add_argument(
        "snapset_name",
        metavar="SNAPSET_NAME",
        type=str,
        action="store",
        help="The name of the snapshot set to create",
    )
    snapset_create_parser.add_argument(
        "sources",
        metavar="SOURCE",
        type=str,
        nargs="+",
        help="A device or mount point path to include in this snapshot set",
    )
    snapset_create_parser.add_argument(
        "-s",
        "--size-policy",
        type=str,
        action="store",
        help="A default size policy for fixed size snapshots",
    )
    snapset_create_parser.add_argument(
        "-b",
        "--bootable",
        action="store_true",
        help="Create a boot entry for this snapshot set",
    )
    snapset_create_parser.add_argument(
        "-r",
        "--revert",
        action="store_true",
        help="Create a revert boot entry for this snapshot set",
    )

    # snapset delete subcommand
    snapset_delete_parser = snapset_subparser.add_parser(
        DELETE_CMD, help="Delete snapshot sets"
    )
    snapset_delete_parser.set_defaults(func=_delete_cmd)
    _add_identifier_args(snapset_delete_parser, snapset=True)

    # snapset rename subcommand
    snapset_rename_parser = snapset_subparser.add_parser(
        RENAME_CMD, help="Rename a snapshot set"
    )
    snapset_rename_parser.set_defaults(func=_rename_cmd)
    snapset_rename_parser.add_argument(
        "old_name",
        metavar="OLD_NAME",
        type=str,
        action="store",
        help="The name of the snapshot set to be renamed",
    )
    snapset_rename_parser.add_argument(
        "new_name",
        metavar="NEW_NAME",
        type=str,
        action="store",
        help="The new name of the snapshot set to be renamed",
    )

    # snapset resize command
    snapset_resize_parser = snapset_subparser.add_parser(
        RESIZE_CMD, help="Resize snapshot sets"
    )
    snapset_resize_parser.set_defaults(func=_resize_cmd)
    _add_identifier_args(snapset_resize_parser, snapset=True)
    snapset_resize_parser.add_argument(
        "sources",
        metavar="SOURCE",
        type=str,
        nargs="*",
        help="A device or mount point path to include in this snapshot set",
    )
    snapset_resize_parser.add_argument(
        "-s",
        "--size-policy",
        type=str,
        action="store",
        help="A default size policy for fixed size snapshots",
    )

    # snapset split command
    snapset_split_parser = snapset_subparser.add_parser(
        SPLIT_CMD, help="Split snapshot sets"
    )
    snapset_split_parser.set_defaults(func=_split_cmd)
    snapset_split_parser.add_argument(
        "name",
        metavar="NAME",
        type=str,
        action="store",
        help="The name of the snapshot set to be split",
    )
    snapset_split_parser.add_argument(
        "new_name",
        metavar="NEW_NAME",
        type=str,
        action="store",
        help="The new name of the new snapshot to split into",
    )
    snapset_split_parser.add_argument(
        "sources",
        metavar="SOURCE",
        type=str,
        nargs="+",
        help="A device or mount point to be split from this snapshot set",
    )

    # snapset prune command
    snapset_prune_parser = snapset_subparser.add_parser(
        PRUNE_CMD, help="Prune snapshots from snapshot sets"
    )
    snapset_prune_parser.set_defaults(func=_prune_cmd)
    snapset_prune_parser.add_argument(
        "name",
        metavar="NAME",
        type=str,
        action="store",
        help="The name of the snapshot set to be split",
    )
    snapset_prune_parser.add_argument(
        "sources",
        metavar="SOURCE",
        type=str,
        nargs="+",
        help="A device or mount point to be pruned from this snapshot set",
    )

    # snapset revert command
    snapset_revert_parser = snapset_subparser.add_parser(
        REVERT_CMD, help="Revert snapshot sets"
    )
    snapset_revert_parser.set_defaults(func=_revert_cmd)
    _add_identifier_args(snapset_revert_parser, snapset=True)

    # snapset activate subcommand
    snapset_activate_parser = snapset_subparser.add_parser(
        ACTIVATE_CMD, help="Activate snapshot sets"
    )
    snapset_activate_parser.set_defaults(func=_activate_cmd)
    _add_identifier_args(snapset_activate_parser, snapset=True)

    # snapset deactivate subcommand
    snapset_deactivate_parser = snapset_subparser.add_parser(
        DEACTIVATE_CMD, help="Deactivate snapshot sets"
    )
    snapset_deactivate_parser.set_defaults(func=_deactivate_cmd)
    _add_identifier_args(snapset_deactivate_parser, snapset=True)

    # snapset autoactivate command
    snapset_autoactivate_parser = snapset_subparser.add_parser(
        AUTOACTIVATE_CMD, help="Set autoactivation status for snapshot sets"
    )
    snapset_autoactivate_parser.set_defaults(func=_autoactivate_cmd)
    _add_identifier_args(snapset_autoactivate_parser, snapset=True)
    _add_autoactivate_args(snapset_autoactivate_parser)

    # snapset list subcommand
    snapset_list_parser = snapset_subparser.add_parser(
        LIST_CMD, help="List snapshot sets"
    )
    snapset_list_parser.set_defaults(func=_list_cmd)
    _add_report_args(snapset_list_parser)
    _add_identifier_args(snapset_list_parser, snapset=True)

    # snapset show subcommand
    snapset_show_parser = snapset_subparser.add_parser(
        SHOW_CMD, help="Display snapshot sets"
    )
    snapset_show_parser.set_defaults(func=_show_cmd)
    _add_identifier_args(snapset_show_parser, snapset=True)
    snapset_show_parser.add_argument(
        "-m",
        "--members",
        action="store_true",
        help="Show snapshots that are part of each snapshot set",
    )
    _add_json_arg(snapset_show_parser)


def _add_snapshot_subparser(type_subparser):
    """
    Add subparser for 'snapshot' commands.

    :param type_subparser: Command type subparser
    """
    # Subparser for snapshot commands
    snapshot_parser = type_subparser.add_parser(SNAPSHOT_TYPE, help="Snapshot commands")
    snapshot_subparser = snapshot_parser.add_subparsers(dest="command")

    # snapshot activate subcommand
    snapshot_activate_parser = snapshot_subparser.add_parser(
        ACTIVATE_CMD, help="Activate snapshots"
    )
    _add_identifier_args(snapshot_activate_parser, snapset=True, snapshot=True)
    snapshot_activate_parser.set_defaults(func=_snapshot_activate_cmd)

    # snapshot deactivate subcommand
    snapshot_deactivate_parser = snapshot_subparser.add_parser(
        DEACTIVATE_CMD, help="Deactivate snapshots"
    )
    _add_identifier_args(snapshot_deactivate_parser, snapset=True, snapshot=True)
    snapshot_deactivate_parser.set_defaults(func=_snapshot_deactivate_cmd)

    # snapshot autoactivate command
    snapshot_autoactivate_parser = snapshot_subparser.add_parser(
        AUTOACTIVATE_CMD, help="Set autoactivation status for snapshots"
    )
    snapshot_autoactivate_parser.set_defaults(func=_snapshot_autoactivate_cmd)
    _add_identifier_args(snapshot_autoactivate_parser, snapset=True, snapshot=True)
    _add_autoactivate_args(snapshot_autoactivate_parser)

    # snapshot list subcommand
    snapshot_list_parser = snapshot_subparser.add_parser(
        LIST_CMD, help="List snapshots"
    )
    snapshot_list_parser.set_defaults(func=_snapshot_list_cmd)
    _add_report_args(snapshot_list_parser)
    _add_identifier_args(snapshot_list_parser, snapset=True, snapshot=True)

    # snapshot show subcommand
    snapshot_show_parser = snapshot_subparser.add_parser(
        SHOW_CMD, help="Display snapshots"
    )
    snapshot_show_parser.set_defaults(func=_snapshot_show_cmd)
    _add_identifier_args(snapshot_show_parser, snapset=True, snapshot=True)
    _add_json_arg(snapshot_show_parser)


def _add_plugin_subparser(type_subparser):
    """
    Add subparser for 'plugin' commands.

    :param type_subparser: Command type subparser
    """
    # Subparser for plugin commands
    plugin_parser = type_subparser.add_parser(PLUGIN_TYPE, help="Plugin commands")
    plugin_subparser = plugin_parser.add_subparsers(dest="command")

    # plugin list subcommand
    plugin_list_parser = plugin_subparser.add_parser(LIST_CMD, help="List plugins")
    plugin_list_parser.set_defaults(func=_plugin_list_cmd)
    _add_report_args(plugin_list_parser)


def main(args):
    """
    Main entry point for snapm.
    """
    parser = ArgumentParser(description="Snapshot Manager", prog=basename(args[0]))

    # Global arguments
    parser.add_argument(
        "-d",
        "--debug",
        metavar="DEBUGOPTS",
        type=str,
        help="A list of debug options to enable",
    )
    parser.add_argument("-v", "--verbose", help="Enable verbose output", action="count")
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        help="Report the version number of snapm",
        version=__version__,
    )
    # Subparser for command type
    type_subparser = parser.add_subparsers(dest="type", help="Command type")

    _add_snapset_subparser(type_subparser)

    _add_snapshot_subparser(type_subparser)

    _add_plugin_subparser(type_subparser)

    cmd_args = parser.parse_args(args[1:])

    status = 1

    try:
        set_debug(cmd_args.debug)
    except ValueError as err:
        print(err)
        parser.print_help()
        return status

    setup_logging(cmd_args)

    _log_debug_command("Parsed %s", " ".join(args[1:]))

    if "func" not in cmd_args:
        parser.print_help()
        return status

    if cmd_args.debug:
        status = cmd_args.func(cmd_args)
    else:
        try:
            status = cmd_args.func(cmd_args)
        # pylint: disable=broad-except
        except KeyboardInterrupt:
            _log_error("Exiting on user cancel")
        except Exception as err:
            _log_error("Command failed: %s", err)

    shutdown_logging()
    return status


# vim: set et ts=4 sw=4 :



# Copyright Red Hat
#
# tests/test_boot.py - Boot support tests
#
# This file is part of the snapm project.
#
# SPDX-License-Identifier: GPL-2.0-only
import unittest
import tempfile
import logging
import os
from subprocess import run
from shutil import rmtree
from uuid import UUID

log = logging.getLogger()

import snapm
import snapm.manager.boot as boot
import snapm.manager
from snapm.manager.plugins import format_snapshot_name, encode_mount_point
import boom

from tests import have_root, BOOT_ROOT_TEST
from ._util import LvmLoopBacked

ETC_FSTAB = "/etc/fstab"
TMP_FSTAB = "/tmp/fstab"

_VAR_TMP = "/var/tmp"


def is_redhat():
    # Exclude CentOS for now until the fix for boom-boot #59 is available
    # in the CentOS repositories.
    return (os.path.exists("/etc/redhat-release")
            and not os.path.exists("/etc/centos-release"))


class BootTestsSimple(unittest.TestCase):
    """
    Test boot helpers
    """

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)

    def tearDown(self):
        log.debug("Tearing down %s", self._testMethodName)

    def test__get_uts_release(self):
        uname_cmd_args = ["uname", "-r"]
        uname_cmd = run(uname_cmd_args, capture_output=True, check=True)
        sys_uts_release = uname_cmd.stdout.decode("utf8").strip()
        self.assertEqual(sys_uts_release, boot._get_uts_release())

    def test__get_machine_id(self):
        with open("/etc/machine-id", "r", encoding="utf8") as id_file:
            sys_machine_id = id_file.read().strip()
        self.assertEqual(sys_machine_id, boot._get_machine_id())


@unittest.skipIf(not have_root(), "requires root privileges")
class BootTests(unittest.TestCase):
    """
    Test boot integration with devices
    """

    volumes = ["root", "home", "var"]
    thin_volumes = ["opt", "srv"]
    boot_volumes = [
        ("root", "/"),
        ("home", "/home"),
        ("var", "/var"),
    ]

    def _set_fstab(self):
        """
        Set up a fake /etc/fstab to be used for the duration of the test run.
        """
        with open(TMP_FSTAB, "w", encoding="utf8") as file:
            for origin, mp in self.boot_volumes:
                file.write(f"/dev/test_vg0/{origin}\t{mp}\text4\tdefaults 0 0\n")
            run(["mount", "--bind", TMP_FSTAB, ETC_FSTAB])

    def _clear_fstab(self):
        """
        Remove the fake /etc/fstab.
        """
        run(["umount", ETC_FSTAB])
        os.unlink(TMP_FSTAB)

    def _populate_boom_root_path(self):
        """
        Populate a mock /boot directory for testing.
        """
        boot_dir = tempfile.mkdtemp("_snapm_boom_dir", dir=_VAR_TMP)
        boom_dir = os.path.join(boot_dir, "boom")
        os.makedirs(os.path.join(boot_dir, "loader", "entries"), exist_ok=True)

        subdirs = ["cache", "hosts", "profiles"]
        for subdir in subdirs:
            os.makedirs(os.path.join(boom_dir, subdir), exist_ok=True)

        boom_conf = [
            "[global]\n",
            f"boot_root = {boot_dir}\n",
            f"boom_root = %(boot_root)s/boom\n",
            "[legacy]\n",
            "enable = False\n",
            "format = grub1\n",
            "sync = False\n",
            "enable = True\n",
            "auto_clean = True\n",
            "cache_path = %(boom_root)/cache\n",
        ]
        with open(os.path.join(boom_dir, "boom.conf"), "w", encoding="utf8") as file:
            file.writelines(boom_conf)

        # Create fake kernel and initramfs for current uname -r
        version = os.uname()[2]
        with open(
            os.path.join(boot_dir, f"vmlinuz-{version}"), "w", encoding="utf8"
        ) as file:
            file.write("test\n")
        with open(
            os.path.join(boot_dir, f"initramfs-{version}.img"), "w", encoding="utf8"
        ) as file:
            file.write("test\n")

        return boot_dir

    def _cleanup_boom_root_path(self, boot_path):
        boom.set_boot_path(BOOT_ROOT_TEST)
        boom.osprofile.load_profiles()
        boom.bootloader.load_entries()
        rmtree(boot_path)

    def setUp(self):
        log.debug("Preparing %s", self._testMethodName)
        boom.set_boot_path("/boot")
        self._lvm = LvmLoopBacked(self.volumes, thin_volumes=self.thin_volumes)
        snapset_name = "bootset0"
        snapset_time = 1707923080
        for origin, mp in self.boot_volumes:
            self._lvm.create_snapshot(
                origin,
                format_snapshot_name(
                    origin, snapset_name, snapset_time, encode_mount_point(mp)
                ),
            )
        self.manager = snapm.manager.Manager()
        self._set_fstab()

    def tearDown(self):
        log.debug("Tearing down %s", self._testMethodName)
        boom.set_boot_path(BOOT_ROOT_TEST)
        self._clear_fstab()
        self._lvm.destroy()

    def test_create_snapshot_boot_entry_no_id(self):
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.create_snapshot_set_boot_entry()

    def test_create_snapshot_revert_entry_no_id(self):
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.create_snapshot_set_revert_entry()

    def test_create_snapshot_boot_entry_bad_name(self):
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.create_snapshot_set_boot_entry(name="bootset1")

    def test_create_snapshot_revert_entry_bad_name(self):
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.create_snapshot_set_revert_entry(name="bootset1")

    def test_create_snapshot_boot_entry(self):
        self.manager.create_snapshot_set_boot_entry(name="bootset0")

        # Clean up boot entry
        self.manager.delete_snapshot_sets(snapm.Selection(name="bootset0"))

    def test_create_snapshot_revert_entry(self):
        self.manager.create_snapshot_set_revert_entry(name="bootset0")

        # Clean up revert entry
        self.manager.delete_snapshot_sets(snapm.Selection(name="bootset0"))

    def test_create_snapshot_boot_entry_bad_uuid(self):
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.create_snapshot_set_boot_entry(
                uuid=UUID("00000000-0000-0000-0000-000000000000")
            )

    def test_create_snapshot_revert_entry_bad_uuid(self):
        with self.assertRaises(snapm.SnapmNotFoundError) as cm:
            self.manager.create_snapshot_set_revert_entry(
                uuid=UUID("00000000-0000-0000-0000-000000000000")
            )

    def test_create_snapshot_boot_entry_uuid(self):
        sset = self.manager.find_snapshot_sets(snapm.Selection(name="bootset0"))[0]
        self.manager.create_snapshot_set_boot_entry(uuid=sset.uuid)

        # Clean up boot entry
        self.manager.delete_snapshot_sets(snapm.Selection(name="bootset0"))

    def test_create_snapshot_revert_entry_uuid(self):
        sset = self.manager.find_snapshot_sets(snapm.Selection(name="bootset0"))[0]
        self.manager.create_snapshot_set_revert_entry(uuid=sset.uuid)

        # Clean up revert entry
        self.manager.delete_snapshot_sets(snapm.Selection(name="bootset0"))

    def test_create_boot_entries_and_discovery(self):
        sset = self.manager.find_snapshot_sets(snapm.Selection(name="bootset0"))[0]
        self.manager.create_snapshot_set_boot_entry(name="bootset0")
        self.manager.create_snapshot_set_revert_entry(name="bootset0")

        # Re-discover snapshot sets w/attached boot entries
        self.manager.discover_snapshot_sets()
        sset = self.manager.find_snapshot_sets(snapm.Selection(name="bootset0"))[0]

        # Validate boot entry
        boot_entry = sset.boot_entry
        for snapshot in sset.snapshots:
            self.assertIn(snapshot.devpath, boot_entry.options)

        # Validate revert entry
        revert_entry = sset.revert_entry
        root_snapshot = sset.snapshot_by_mount_point("/")
        self.assertIn(root_snapshot.origin, revert_entry.options)

        # Clean up boot entries
        self.manager.delete_snapshot_sets(snapm.Selection(name="bootset0"))

    @unittest.skipIf(not is_redhat(), "profile auto-creation not supported")
    def test_auto_profile_create_boot_entry(self):
        boot_dir = self._populate_boom_root_path()
        boom.set_boot_path(boot_dir)
        boom.osprofile.load_profiles()
        boom.bootloader.load_entries()
        self.addCleanup(self._cleanup_boom_root_path, boot_dir)
        sset = self.manager.find_snapshot_sets(snapm.Selection(name="bootset0"))[0]
        self.manager.create_snapshot_set_boot_entry(uuid=sset.uuid)

        # Clean up boot entry
        self.manager.delete_snapshot_sets(snapm.Selection(name="bootset0"))
